# Copyright 2018 Verizon Wireless
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys
from unittest import mock
import urllib.error

from glance_store import exceptions as store_exceptions
from oslo_config import cfg
from oslo_utils import units
import taskflow

import glance.async_.flows.api_image_import as import_flow
from glance.common import exception
from glance.common.scripts.image_import import main as image_import
from glance.common.scripts import utils as script_utils
from glance import context
from glance.domain import ExtraProperties
from glance import gateway
import glance.tests.utils as test_utils

from cursive import exception as cursive_exception

CONF = cfg.CONF

TASK_TYPE = 'api_image_import'
TASK_ID1 = 'dbbe7231-020f-4311-87e1-5aaa6da56c02'
IMAGE_ID1 = '41f5b3b0-f54c-4cef-bd45-ce3e376a142f'
UUID1 = 'c80a1a6c-bd1f-41c5-90ee-81afedb1d58d'
TENANT1 = '6838eb7b-6ded-434a-882c-b344c77fe8df'


class TestApiImageImportTask(test_utils.BaseTestCase):

    def setUp(self):
        super(TestApiImageImportTask, self).setUp()

        self.wd_task_input = {
            "import_req": {
                "method": {
                    "name": "web-download",
                    "uri": "http://example.com/image.browncow"
                }
            }
        }

        self.gd_task_input = {
            "import_req": {
                "method": {
                    "name": "glance-direct"
                }
            }
        }

        self.mock_task_repo = mock.MagicMock()
        self.mock_image_repo = mock.MagicMock()
        self.mock_image = self.mock_image_repo.get.return_value
        self.mock_image.extra_properties = {
            'os_glance_import_task': TASK_ID1,
            'os_glance_stage_host': 'http://glance2',
        }

    @mock.patch('glance.async_.flows.api_image_import._VerifyStaging.__init__')
    @mock.patch('taskflow.patterns.linear_flow.Flow.add')
    @mock.patch('taskflow.patterns.linear_flow.__init__')
    def _pass_uri(self, mock_lf_init, mock_flow_add, mock_VS_init,
                  uri, file_uri, import_req):
        flow_kwargs = {"task_id": TASK_ID1,
                       "task_type": TASK_TYPE,
                       "task_repo": self.mock_task_repo,
                       "image_repo": self.mock_image_repo,
                       "image_id": IMAGE_ID1,
                       "context": mock.MagicMock(),
                       "import_req": import_req}

        mock_lf_init.return_value = None
        mock_VS_init.return_value = None

        self.config(node_staging_uri=uri)
        import_flow.get_flow(**flow_kwargs)
        mock_VS_init.assert_called_with(TASK_ID1, TASK_TYPE,
                                        self.mock_task_repo,
                                        file_uri)

    def test_get_flow_handles_node_uri_with_ending_slash(self):
        test_uri = 'file:///some/where/'
        expected_uri = '{0}{1}'.format(test_uri, IMAGE_ID1)
        self._pass_uri(uri=test_uri, file_uri=expected_uri,
                       import_req=self.gd_task_input['import_req'])
        self._pass_uri(uri=test_uri, file_uri=expected_uri,
                       import_req=self.wd_task_input['import_req'])

    def test_get_flow_handles_node_uri_without_ending_slash(self):
        test_uri = 'file:///some/where'
        expected_uri = '{0}/{1}'.format(test_uri, IMAGE_ID1)
        self._pass_uri(uri=test_uri, file_uri=expected_uri,
                       import_req=self.wd_task_input['import_req'])
        self._pass_uri(uri=test_uri, file_uri=expected_uri,
                       import_req=self.gd_task_input['import_req'])

    def test_get_flow_pops_stage_host(self):
        import_flow.get_flow(task_id=TASK_ID1, task_type=TASK_TYPE,
                             task_repo=self.mock_task_repo,
                             image_repo=self.mock_image_repo,
                             image_id=IMAGE_ID1,
                             context=mock.MagicMock(),
                             import_req=self.gd_task_input['import_req'])
        self.assertNotIn('os_glance_stage_host',
                         self.mock_image.extra_properties)
        self.assertIn('os_glance_import_task',
                      self.mock_image.extra_properties)

    def test_assert_quota_no_task(self):
        ignored = mock.MagicMock()
        task_repo = mock.MagicMock()
        task_repo.get.return_value = None
        task_id = 'some-task'
        enforce_fn = mock.MagicMock()
        enforce_fn.side_effect = exception.LimitExceeded
        with mock.patch.object(import_flow, 'LOG') as mock_log:
            self.assertRaises(exception.LimitExceeded,
                              import_flow.assert_quota,
                              ignored, task_repo, task_id,
                              [], ignored, enforce_fn)
        task_repo.get.assert_called_once_with('some-task')
        # Make sure we logged instead of crashed if no task was found
        mock_log.error.assert_called_once_with('Failed to find task %r to '
                                               'update after quota failure',
                                               'some-task')
        task_repo.save.assert_not_called()

    def test_assert_quota(self):
        ignored = mock.MagicMock()
        task_repo = mock.MagicMock()
        task_id = 'some-task'
        enforce_fn = mock.MagicMock()
        enforce_fn.side_effect = exception.LimitExceeded
        wrapper = mock.MagicMock()
        action = wrapper.__enter__.return_value
        action.image_status = 'importing'
        self.assertRaises(exception.LimitExceeded,
                          import_flow.assert_quota,
                          ignored, task_repo, task_id,
                          ['store1'], wrapper, enforce_fn)
        action.remove_importing_stores.assert_called_once_with(['store1'])
        action.set_image_attribute.assert_called_once_with(status='queued')
        task_repo.get.assert_called_once_with('some-task')
        task_repo.save.assert_called_once_with(task_repo.get.return_value)

    def test_assert_quota_copy(self):
        ignored = mock.MagicMock()
        task_repo = mock.MagicMock()
        task_id = 'some-task'
        enforce_fn = mock.MagicMock()
        enforce_fn.side_effect = exception.LimitExceeded
        wrapper = mock.MagicMock()
        action = wrapper.__enter__.return_value
        action.image_status = 'active'
        self.assertRaises(exception.LimitExceeded,
                          import_flow.assert_quota,
                          ignored, task_repo, task_id,
                          ['store1'], wrapper, enforce_fn)
        action.remove_importing_stores.assert_called_once_with(['store1'])
        action.set_image_attribute.assert_not_called()
        task_repo.get.assert_called_once_with('some-task')
        task_repo.save.assert_called_once_with(task_repo.get.return_value)


class TestImageLock(test_utils.BaseTestCase):
    def setUp(self):
        super(TestImageLock, self).setUp()
        self.img_repo = mock.MagicMock()

    @mock.patch('glance.async_.flows.api_image_import.LOG')
    def test_execute_confirms_lock(self, mock_log):
        self.img_repo.get.return_value.extra_properties = {
            'os_glance_import_task': TASK_ID1}
        wrapper = import_flow.ImportActionWrapper(self.img_repo, IMAGE_ID1,
                                                  TASK_ID1)
        imagelock = import_flow._ImageLock(TASK_ID1, TASK_TYPE, wrapper)
        imagelock.execute()
        mock_log.debug.assert_called_once_with('Image %(image)s import task '
                                               '%(task)s lock confirmed',
                                               {'image': IMAGE_ID1,
                                                'task': TASK_ID1})

    @mock.patch('glance.async_.flows.api_image_import.LOG')
    def test_execute_confirms_lock_not_held(self, mock_log):
        wrapper = import_flow.ImportActionWrapper(self.img_repo, IMAGE_ID1,
                                                  TASK_ID1)
        imagelock = import_flow._ImageLock(TASK_ID1, TASK_TYPE, wrapper)
        self.assertRaises(exception.TaskAbortedError,
                          imagelock.execute)

    @mock.patch('glance.async_.flows.api_image_import.LOG')
    def test_revert_drops_lock(self, mock_log):
        wrapper = import_flow.ImportActionWrapper(self.img_repo, IMAGE_ID1,
                                                  TASK_ID1)
        imagelock = import_flow._ImageLock(TASK_ID1, TASK_TYPE, wrapper)
        with mock.patch.object(wrapper, 'drop_lock_for_task') as mock_drop:
            imagelock.revert(None)
            mock_drop.assert_called_once_with()
        mock_log.debug.assert_called_once_with('Image %(image)s import task '
                                               '%(task)s dropped its lock '
                                               'after failure',
                                               {'image': IMAGE_ID1,
                                                'task': TASK_ID1})

    @mock.patch('glance.async_.flows.api_image_import.LOG')
    def test_revert_drops_lock_missing(self, mock_log):
        wrapper = import_flow.ImportActionWrapper(self.img_repo, IMAGE_ID1,
                                                  TASK_ID1)
        imagelock = import_flow._ImageLock(TASK_ID1, TASK_TYPE, wrapper)
        with mock.patch.object(wrapper, 'drop_lock_for_task') as mock_drop:
            mock_drop.side_effect = exception.NotFound()
            imagelock.revert(None)
        mock_log.warning.assert_called_once_with('Image %(image)s import task '
                                                 '%(task)s lost its lock '
                                                 'during execution!',
                                                 {'image': IMAGE_ID1,
                                                  'task': TASK_ID1})


class TestImportToStoreTask(test_utils.BaseTestCase):

    def setUp(self):
        super(TestImportToStoreTask, self).setUp()
        self.gateway = gateway.Gateway()
        self.context = context.RequestContext(user_id=TENANT1,
                                              project_id=TENANT1,
                                              overwrite=False)
        self.img_factory = self.gateway.get_image_factory(self.context)

    def test_execute(self):
        wrapper = mock.MagicMock()
        action = mock.MagicMock()
        task_repo = mock.MagicMock()
        wrapper.__enter__.return_value = action
        image_import = import_flow._ImportToStore(TASK_ID1, TASK_TYPE,
                                                  task_repo, wrapper,
                                                  "http://url",
                                                  "store1", False,
                                                  True)
        # Assert file_path is honored
        with mock.patch.object(image_import, '_execute') as mock_execute:
            image_import.execute(mock.sentinel.path)
            mock_execute.assert_called_once_with(action, mock.sentinel.path)

        # Assert file_path is optional
        with mock.patch.object(image_import, '_execute') as mock_execute:
            image_import.execute()
            mock_execute.assert_called_once_with(action, None)

    def test_execute_body_with_store(self):
        image = mock.MagicMock()
        img_repo = mock.MagicMock()
        img_repo.get.return_value = image
        task_repo = mock.MagicMock()
        wrapper = import_flow.ImportActionWrapper(img_repo, IMAGE_ID1,
                                                  TASK_ID1)
        image_import = import_flow._ImportToStore(TASK_ID1, TASK_TYPE,
                                                  task_repo, wrapper,
                                                  "http://url",
                                                  "store1", False,
                                                  True)
        action = mock.MagicMock()
        image_import._execute(action, mock.sentinel.path)
        action.set_image_data.assert_called_once_with(
            mock.sentinel.path,
            TASK_ID1, backend='store1',
            set_active=True,
            callback=image_import._status_callback)
        action.remove_importing_stores(['store1'])

    def test_execute_body_with_store_no_path(self):
        image = mock.MagicMock()
        img_repo = mock.MagicMock()
        img_repo.get.return_value = image
        task_repo = mock.MagicMock()
        wrapper = import_flow.ImportActionWrapper(img_repo, IMAGE_ID1,
                                                  TASK_ID1)
        image_import = import_flow._ImportToStore(TASK_ID1, TASK_TYPE,
                                                  task_repo, wrapper,
                                                  "http://url",
                                                  "store1", False,
                                                  True)
        action = mock.MagicMock()
        image_import._execute(action, None)
        action.set_image_data.assert_called_once_with(
            'http://url',
            TASK_ID1, backend='store1',
            set_active=True,
            callback=image_import._status_callback)
        action.remove_importing_stores(['store1'])

    def test_execute_body_without_store(self):
        image = mock.MagicMock()
        img_repo = mock.MagicMock()
        img_repo.get.return_value = image
        task_repo = mock.MagicMock()
        wrapper = import_flow.ImportActionWrapper(img_repo, IMAGE_ID1,
                                                  TASK_ID1)
        image_import = import_flow._ImportToStore(TASK_ID1, TASK_TYPE,
                                                  task_repo, wrapper,
                                                  "http://url",
                                                  None, False,
                                                  True)
        action = mock.MagicMock()
        image_import._execute(action, mock.sentinel.path)
        action.set_image_data.assert_called_once_with(
            mock.sentinel.path,
            TASK_ID1, backend=None,
            set_active=True,
            callback=image_import._status_callback)
        action.remove_importing_stores.assert_not_called()

    @mock.patch('glance.async_.flows.api_image_import.LOG.debug')
    @mock.patch('oslo_utils.timeutils.now')
    def test_status_callback_limits_rate(self, mock_now, mock_log):
        img_repo = mock.MagicMock()
        task_repo = mock.MagicMock()
        task_repo.get.return_value.status = 'processing'
        wrapper = import_flow.ImportActionWrapper(img_repo, IMAGE_ID1,
                                                  TASK_ID1)
        image_import = import_flow._ImportToStore(TASK_ID1, TASK_TYPE,
                                                  task_repo, wrapper,
                                                  "http://url",
                                                  None, False,
                                                  True)

        expected_calls = []
        log_call = mock.call('Image import %(image_id)s copied %(copied)i MiB',
                             {'image_id': IMAGE_ID1,
                              'copied': 0})
        action = mock.MagicMock(image_id=IMAGE_ID1)

        mock_now.return_value = 1000
        image_import._status_callback(action, 32, 32)
        # First call will emit immediately because we only ran __init__
        # which sets the last status to zero
        expected_calls.append(log_call)
        mock_log.assert_has_calls(expected_calls)

        image_import._status_callback(action, 32, 64)
        # Second call will not emit any other logs because no time
        # has passed
        mock_log.assert_has_calls(expected_calls)

        mock_now.return_value += 190
        image_import._status_callback(action, 32, 96)
        # Third call will not emit any other logs because not enough
        # time has passed
        mock_log.assert_has_calls(expected_calls)

        mock_now.return_value += 300
        image_import._status_callback(action, 32, 128)
        # Fourth call will emit because we crossed five minutes
        expected_calls.append(log_call)
        mock_log.assert_has_calls(expected_calls)

        mock_now.return_value += 150
        image_import._status_callback(action, 32, 128)
        # Fifth call will not emit any other logs because not enough
        # time has passed
        mock_log.assert_has_calls(expected_calls)

        mock_now.return_value += 3600
        image_import._status_callback(action, 32, 128)
        # Sixth call will emit because we crossed five minutes
        expected_calls.append(log_call)
        mock_log.assert_has_calls(expected_calls)

    def test_raises_when_image_deleted(self):
        img_repo = mock.MagicMock()
        task_repo = mock.MagicMock()
        wrapper = import_flow.ImportActionWrapper(img_repo, IMAGE_ID1,
                                                  TASK_ID1)
        image_import = import_flow._ImportToStore(TASK_ID1, TASK_TYPE,
                                                  task_repo, wrapper,
                                                  "http://url",
                                                  "store1", False,
                                                  True)
        image = self.img_factory.new_image(image_id=UUID1)
        image.status = "deleted"
        img_repo.get.return_value = image
        self.assertRaises(exception.ImportTaskError, image_import.execute)

    @mock.patch("glance.async_.flows.api_image_import.image_import")
    def test_remove_store_from_property(self, mock_import):
        img_repo = mock.MagicMock()
        task_repo = mock.MagicMock()
        wrapper = import_flow.ImportActionWrapper(img_repo, IMAGE_ID1,
                                                  TASK_ID1)
        image_import = import_flow._ImportToStore(TASK_ID1, TASK_TYPE,
                                                  task_repo, wrapper,
                                                  "http://url",
                                                  "store1", True,
                                                  True)
        extra_properties = {"os_glance_importing_to_stores": "store1,store2",
                            "os_glance_import_task": TASK_ID1}
        image = self.img_factory.new_image(image_id=UUID1,
                                           extra_properties=extra_properties)
        img_repo.get.return_value = image
        image_import.execute()
        self.assertEqual(
            image.extra_properties['os_glance_importing_to_stores'], "store2")

    def test_revert_updates_status_keys(self):
        img_repo = mock.MagicMock()
        task_repo = mock.MagicMock()
        wrapper = import_flow.ImportActionWrapper(img_repo, IMAGE_ID1,
                                                  TASK_ID1)
        image_import = import_flow._ImportToStore(TASK_ID1, TASK_TYPE,
                                                  task_repo, wrapper,
                                                  "http://url",
                                                  "store1", True,
                                                  True)
        extra_properties = {"os_glance_importing_to_stores": "store1,store2",
                            "os_glance_import_task": TASK_ID1}
        image = self.img_factory.new_image(image_id=UUID1,
                                           extra_properties=extra_properties)
        img_repo.get.return_value = image

        fail_key = 'os_glance_failed_import'
        pend_key = 'os_glance_importing_to_stores'

        image_import.revert(None)
        self.assertEqual('store2', image.extra_properties[pend_key])

        try:
            raise Exception('foo')
        except Exception:
            fake_exc_info = sys.exc_info()

        extra_properties = {"os_glance_importing_to_stores": "store1,store2"}
        image_import.revert(taskflow.types.failure.Failure(fake_exc_info))
        self.assertEqual('store2', image.extra_properties[pend_key])
        self.assertEqual('store1', image.extra_properties[fail_key])

    @mock.patch("glance.async_.flows.api_image_import.image_import")
    def test_raises_when_all_stores_must_succeed(self, mock_import):
        img_repo = mock.MagicMock()
        task_repo = mock.MagicMock()
        wrapper = import_flow.ImportActionWrapper(img_repo, IMAGE_ID1,
                                                  TASK_ID1)
        image_import = import_flow._ImportToStore(TASK_ID1, TASK_TYPE,
                                                  task_repo, wrapper,
                                                  "http://url",
                                                  "store1", True,
                                                  True)
        extra_properties = {'os_glance_import_task': TASK_ID1}
        image = self.img_factory.new_image(image_id=UUID1,
                                           extra_properties=extra_properties)
        img_repo.get.return_value = image
        mock_import.set_image_data.side_effect = \
            cursive_exception.SignatureVerificationError(
                "Signature verification failed")
        self.assertRaises(cursive_exception.SignatureVerificationError,
                          image_import.execute)

    @mock.patch("glance.async_.flows.api_image_import.image_import")
    def test_doesnt_raise_when_not_all_stores_must_succeed(self, mock_import):
        img_repo = mock.MagicMock()
        task_repo = mock.MagicMock()
        wrapper = import_flow.ImportActionWrapper(img_repo, IMAGE_ID1,
                                                  TASK_ID1)
        image_import = import_flow._ImportToStore(TASK_ID1, TASK_TYPE,
                                                  task_repo, wrapper,
                                                  "http://url",
                                                  "store1", False,
                                                  True)
        extra_properties = {'os_glance_import_task': TASK_ID1}
        image = self.img_factory.new_image(image_id=UUID1,
                                           extra_properties=extra_properties)
        img_repo.get.return_value = image
        mock_import.set_image_data.side_effect = \
            cursive_exception.SignatureVerificationError(
                "Signature verification failed")
        try:
            image_import.execute()
            self.assertEqual(image.extra_properties['os_glance_failed_import'],
                             "store1")
        except cursive_exception.SignatureVerificationError:
            self.fail("Exception shouldn't be raised")

    @mock.patch('glance.common.scripts.utils.get_task')
    def test_status_callback_updates_task_message(self, mock_get):
        task_repo = mock.MagicMock()
        image_import = import_flow._ImportToStore(TASK_ID1, TASK_TYPE,
                                                  task_repo, mock.MagicMock(),
                                                  "http://url",
                                                  "store1", False,
                                                  True)
        task = mock.MagicMock()
        task.status = 'processing'
        mock_get.return_value = task
        action = mock.MagicMock()
        image_import._status_callback(action, 128, 256 * units.Mi)
        mock_get.assert_called_once_with(task_repo, TASK_ID1)
        task_repo.save.assert_called_once_with(task)
        self.assertEqual(_('Copied %i MiB' % 256), task.message)

    @mock.patch('glance.common.scripts.utils.get_task')
    def test_status_aborts_missing_task(self, mock_get):
        task_repo = mock.MagicMock()
        image_import = import_flow._ImportToStore(TASK_ID1, TASK_TYPE,
                                                  task_repo, mock.MagicMock(),
                                                  "http://url",
                                                  "store1", False,
                                                  True)
        mock_get.return_value = None
        action = mock.MagicMock()
        self.assertRaises(exception.TaskNotFound,
                          image_import._status_callback,
                          action, 128, 256 * units.Mi)
        mock_get.assert_called_once_with(task_repo, TASK_ID1)
        task_repo.save.assert_not_called()

    @mock.patch('glance.common.scripts.utils.get_task')
    def test_status_aborts_invalid_task_state(self, mock_get):
        task_repo = mock.MagicMock()
        image_import = import_flow._ImportToStore(TASK_ID1, TASK_TYPE,
                                                  task_repo, mock.MagicMock(),
                                                  "http://url",
                                                  "store1", False,
                                                  True)
        task = mock.MagicMock()
        task.status = 'failed'
        mock_get.return_value = task
        action = mock.MagicMock()
        self.assertRaises(exception.TaskAbortedError,
                          image_import._status_callback,
                          action, 128, 256 * units.Mi)
        mock_get.assert_called_once_with(task_repo, TASK_ID1)
        task_repo.save.assert_not_called()


class TestDeleteFromFS(test_utils.BaseTestCase):
    def test_delete_with_backends_deletes(self):
        task = import_flow._DeleteFromFS(TASK_ID1, TASK_TYPE)
        self.config(enabled_backends='file:foo')
        with mock.patch.object(import_flow.store_api, 'delete') as mock_del:
            task.execute(mock.sentinel.path)
            mock_del.assert_called_once_with(
                mock.sentinel.path,
                'os_glance_staging_store')

    def test_delete_with_backends_delete_fails(self):
        self.config(enabled_backends='file:foo')
        task = import_flow._DeleteFromFS(TASK_ID1, TASK_TYPE)
        with mock.patch.object(import_flow.store_api, 'delete') as mock_del:
            mock_del.side_effect = store_exceptions.NotFound(image=IMAGE_ID1,
                                                             message='Testing')
            # If we didn't swallow this we would explode here
            task.execute(mock.sentinel.path)
            mock_del.assert_called_once_with(
                mock.sentinel.path,
                'os_glance_staging_store')

            # Raise something unexpected and make sure it bubbles up
            mock_del.side_effect = RuntimeError
            self.assertRaises(RuntimeError,
                              task.execute, mock.sentinel.path)

    @mock.patch('os.path.exists')
    @mock.patch('os.unlink')
    def test_delete_without_backends_exists(self, mock_unlink, mock_exists):
        mock_exists.return_value = True
        task = import_flow._DeleteFromFS(TASK_ID1, TASK_TYPE)
        task.execute('1234567foo')
        # FIXME(danms): I have no idea why the code arbitrarily snips
        # the first seven characters from the path. Need a comment or
        # *something*.
        mock_unlink.assert_called_once_with('foo')

        mock_unlink.reset_mock()
        mock_unlink.side_effect = OSError(123, 'failed')
        # Make sure we swallow the OSError and don't explode
        task.execute('1234567foo')

    @mock.patch('os.path.exists')
    @mock.patch('os.unlink')
    def test_delete_without_backends_missing(self, mock_unlink, mock_exists):
        mock_exists.return_value = False
        task = import_flow._DeleteFromFS(TASK_ID1, TASK_TYPE)
        task.execute('foo')
        mock_unlink.assert_not_called()


class TestImportCopyImageTask(test_utils.BaseTestCase):

    def setUp(self):
        super(TestImportCopyImageTask, self).setUp()

        self.context = context.RequestContext(user_id=TENANT1,
                                              project_id=TENANT1,
                                              overwrite=False)

    @mock.patch("glance.async_.flows.api_image_import.image_import")
    @mock.patch('glance_store.get_store_from_store_identifier')
    def test_init_copy_flow_as_non_owner(self, mock_gs, mock_import):
        img_repo = mock.MagicMock()
        admin_repo = mock.MagicMock()

        fake_req = {"method": {"name": "copy-image"},
                    "backend": ['cheap']}

        fake_img = mock.MagicMock()
        fake_img.id = IMAGE_ID1
        fake_img.status = 'active'
        fake_img.extra_properties = {'os_glance_import_task': TASK_ID1}
        admin_repo.get.return_value = fake_img

        import_flow.get_flow(task_id=TASK_ID1,
                             task_type=TASK_TYPE,
                             task_repo=mock.MagicMock(),
                             image_repo=img_repo,
                             admin_repo=admin_repo,
                             image_id=IMAGE_ID1,
                             import_req=fake_req,
                             context=self.context,
                             backend=['cheap'])

        # Assert that we saved the image with the admin repo instead of the
        # user-context one at the end of get_flow() when we initialize the
        # parameters.
        admin_repo.save.assert_called_once_with(fake_img, 'active')
        img_repo.save.assert_not_called()


class TestVerifyImageStateTask(test_utils.BaseTestCase):
    def test_verify_active_status(self):
        fake_img = mock.MagicMock(status='active',
                                  extra_properties={
                                      'os_glance_import_task': TASK_ID1})
        mock_repo = mock.MagicMock()
        mock_repo.get.return_value = fake_img
        wrapper = import_flow.ImportActionWrapper(mock_repo, IMAGE_ID1,
                                                  TASK_ID1)

        task = import_flow._VerifyImageState(TASK_ID1, TASK_TYPE,
                                             wrapper, 'anything!')

        task.execute()

        fake_img.status = 'importing'
        self.assertRaises(import_flow._NoStoresSucceeded,
                          task.execute)

    def test_revert_copy_status_unchanged(self):
        wrapper = mock.MagicMock()
        task = import_flow._VerifyImageState(TASK_ID1, TASK_TYPE,
                                             wrapper, 'copy-image')
        task.revert(mock.sentinel.result)

        # If we are doing copy-image, no state update should be made
        wrapper.__enter__.return_value.set_image_attribute.assert_not_called()

    def test_reverts_state_nocopy(self):
        wrapper = mock.MagicMock()
        task = import_flow._VerifyImageState(TASK_ID1, TASK_TYPE,
                                             wrapper, 'glance-direct')
        task.revert(mock.sentinel.result)

        # Except for copy-image, image state should revert to queued
        action = wrapper.__enter__.return_value
        action.set_image_attribute.assert_called_once_with(status='queued')


class TestImportActionWrapper(test_utils.BaseTestCase):
    def test_wrapper_success(self):
        mock_repo = mock.MagicMock()
        mock_repo.get.return_value.extra_properties = {
            'os_glance_import_task': TASK_ID1}
        wrapper = import_flow.ImportActionWrapper(mock_repo, IMAGE_ID1,
                                                  TASK_ID1)
        with wrapper as action:
            self.assertIsInstance(action, import_flow._ImportActions)
        mock_repo.get.assert_has_calls([mock.call(IMAGE_ID1),
                                        mock.call(IMAGE_ID1)])
        mock_repo.save.assert_called_once_with(
            mock_repo.get.return_value,
            mock_repo.get.return_value.status)

    def test_wrapper_failure(self):
        mock_repo = mock.MagicMock()
        mock_repo.get.return_value.extra_properties = {
            'os_glance_import_task': TASK_ID1}
        wrapper = import_flow.ImportActionWrapper(mock_repo, IMAGE_ID1,
                                                  TASK_ID1)

        class SpecificError(Exception):
            pass

        try:
            with wrapper:
                raise SpecificError('some failure')
        except SpecificError:
            # NOTE(danms): Make sure we only caught the test exception
            # and aren't hiding anything else
            pass

        mock_repo.get.assert_called_once_with(IMAGE_ID1)
        mock_repo.save.assert_not_called()

    @mock.patch.object(import_flow, 'LOG')
    def test_wrapper_logs_status(self, mock_log):
        mock_repo = mock.MagicMock()
        mock_image = mock_repo.get.return_value
        mock_image.extra_properties = {'os_glance_import_task': TASK_ID1}
        wrapper = import_flow.ImportActionWrapper(mock_repo, IMAGE_ID1,
                                                  TASK_ID1)

        mock_image.status = 'foo'
        with wrapper as action:
            action.set_image_attribute(status='bar')

        mock_log.debug.assert_called_once_with(
            'Image %(image_id)s status changing from '
            '%(old_status)s to %(new_status)s',
            {'image_id': IMAGE_ID1,
             'old_status': 'foo',
             'new_status': 'bar'})
        self.assertEqual('bar', mock_image.status)

    def test_image_id_property(self):
        mock_repo = mock.MagicMock()
        wrapper = import_flow.ImportActionWrapper(mock_repo, IMAGE_ID1,
                                                  TASK_ID1)
        self.assertEqual(IMAGE_ID1, wrapper.image_id)

    def test_set_image_attribute(self):
        mock_repo = mock.MagicMock()
        mock_image = mock_repo.get.return_value
        mock_image.extra_properties = {'os_glance_import_task': TASK_ID1}
        mock_image.status = 'bar'
        wrapper = import_flow.ImportActionWrapper(mock_repo, IMAGE_ID1,
                                                  TASK_ID1)
        with wrapper as action:
            action.set_image_attribute(status='foo', virtual_size=123,
                                       size=64)
        mock_repo.save.assert_called_once_with(mock_image, 'bar')
        self.assertEqual('foo', mock_image.status)
        self.assertEqual(123, mock_image.virtual_size)
        self.assertEqual(64, mock_image.size)

    def test_set_image_attribute_disallowed(self):
        mock_repo = mock.MagicMock()
        mock_image = mock_repo.get.return_value
        mock_image.extra_properties = {'os_glance_import_task': TASK_ID1}
        mock_image.status = 'bar'
        wrapper = import_flow.ImportActionWrapper(mock_repo, IMAGE_ID1,
                                                  TASK_ID1)
        with wrapper as action:
            self.assertRaises(AttributeError,
                              action.set_image_attribute, id='foo')

    @mock.patch.object(import_flow, 'LOG')
    def test_set_image_extra_properties(self, mock_log):
        mock_repo = mock.MagicMock()
        mock_image = mock_repo.get.return_value
        mock_image.image_id = IMAGE_ID1
        mock_image.extra_properties = {'os_glance_import_task': TASK_ID1}
        mock_image.status = 'bar'
        wrapper = import_flow.ImportActionWrapper(mock_repo, IMAGE_ID1,
                                                  TASK_ID1)
        # One banned property
        with wrapper as action:
            action.set_image_extra_properties({'os_glance_foo': 'bar'})
        self.assertEqual({'os_glance_import_task': TASK_ID1},
                         mock_image.extra_properties)

        mock_log.warning.assert_called()
        mock_log.warning.reset_mock()

        # Two banned properties
        with wrapper as action:
            action.set_image_extra_properties({'os_glance_foo': 'bar',
                                               'os_glance_baz': 'bat'})
        self.assertEqual({'os_glance_import_task': TASK_ID1},
                         mock_image.extra_properties)

        mock_log.warning.assert_called()
        mock_log.warning.reset_mock()

        # One banned and one allowed property
        with wrapper as action:
            action.set_image_extra_properties({'foo': 'bar',
                                               'os_glance_foo': 'baz'})
        self.assertEqual({'foo': 'bar',
                          'os_glance_import_task': TASK_ID1},
                         mock_image.extra_properties)

        mock_log.warning.assert_called_once_with(
            'Dropping %(key)s=%(val)s during metadata injection for %(image)s',
            {'key': 'os_glance_foo', 'val': 'baz',
             'image': IMAGE_ID1})

    def test_image_size(self):
        mock_repo = mock.MagicMock()
        mock_image = mock_repo.get.return_value
        mock_image.image_id = IMAGE_ID1
        mock_image.extra_properties = {'os_glance_import_task': TASK_ID1}
        mock_image.size = 123
        wrapper = import_flow.ImportActionWrapper(mock_repo, IMAGE_ID1,
                                                  TASK_ID1)
        with wrapper as action:
            self.assertEqual(123, action.image_size)

    def test_image_locations(self):
        mock_repo = mock.MagicMock()
        mock_image = mock_repo.get.return_value
        mock_image.image_id = IMAGE_ID1
        mock_image.extra_properties = {'os_glance_import_task': TASK_ID1}
        mock_image.locations = {'some': {'complex': ['structure']}}
        wrapper = import_flow.ImportActionWrapper(mock_repo, IMAGE_ID1,
                                                  TASK_ID1)
        with wrapper as action:
            self.assertEqual({'some': {'complex': ['structure']}},
                             action.image_locations)
            # Mutate our copy
            action.image_locations['foo'] = 'bar'

        # Make sure we did not mutate the image itself
        self.assertEqual({'some': {'complex': ['structure']}},
                         mock_image.locations)

    def test_drop_lock_for_task(self):
        mock_repo = mock.MagicMock()
        mock_repo.get.return_value.extra_properties = {
            'os_glance_import_task': TASK_ID1}
        wrapper = import_flow.ImportActionWrapper(mock_repo, IMAGE_ID1,
                                                  TASK_ID1)
        wrapper.drop_lock_for_task()
        mock_repo.delete_property_atomic.assert_called_once_with(
            mock_repo.get.return_value, 'os_glance_import_task', TASK_ID1)

    def test_assert_task_lock(self):
        mock_repo = mock.MagicMock()
        mock_repo.get.return_value.extra_properties = {
            'os_glance_import_task': TASK_ID1}
        wrapper = import_flow.ImportActionWrapper(mock_repo, IMAGE_ID1,
                                                  TASK_ID1)
        wrapper.assert_task_lock()

        # Try again with a different task ID and it should fail
        wrapper = import_flow.ImportActionWrapper(mock_repo, IMAGE_ID1,
                                                  'foo')
        self.assertRaises(exception.TaskAbortedError,
                          wrapper.assert_task_lock)

    def _grab_image(self, wrapper):
        with wrapper:
            pass

    @mock.patch.object(import_flow, 'LOG')
    def test_check_task_lock(self, mock_log):
        mock_repo = mock.MagicMock()
        wrapper = import_flow.ImportActionWrapper(mock_repo, IMAGE_ID1,
                                                  TASK_ID1)
        image = mock.MagicMock(image_id=IMAGE_ID1)
        image.extra_properties = {'os_glance_import_task': TASK_ID1}
        mock_repo.get.return_value = image
        self._grab_image(wrapper)
        mock_log.error.assert_not_called()

        image.extra_properties['os_glance_import_task'] = 'somethingelse'
        self.assertRaises(exception.TaskAbortedError,
                          self._grab_image, wrapper)
        mock_log.error.assert_called_once_with(
            'Image %(image)s import task %(task)s attempted to take action on '
            'image, but other task %(other)s holds the lock; Aborting.',
            {'image': image.image_id,
             'task': TASK_ID1,
             'other': 'somethingelse'})


class TestImportActions(test_utils.BaseTestCase):
    def setUp(self):
        super(TestImportActions, self).setUp()
        self.image = mock.MagicMock()
        self.image.image_id = IMAGE_ID1
        self.image.status = 'active'
        self.image.disk_format = 'raw'
        self.image.container_format = 'bare'
        self.image.extra_properties = ExtraProperties({'speed': '88mph'})
        self.image.checksum = mock.sentinel.checksum
        self.image.os_hash_algo = mock.sentinel.hash_algo
        self.image.os_hash_value = mock.sentinel.hash_value
        self.image.size = mock.sentinel.size
        self.actions = import_flow._ImportActions(self.image)

    def test_image_property_proxies(self):
        self.assertEqual(IMAGE_ID1, self.actions.image_id)
        self.assertEqual('active', self.actions.image_status)
        self.assertEqual('raw', self.actions.image_disk_format)
        self.assertEqual('bare', self.actions.image_container_format)
        self.assertEqual({'speed': '88mph'},
                         self.actions.image_extra_properties)

    def test_merge_store_list(self):
        # Addition with no existing property works
        self.actions.merge_store_list('stores', ['foo', 'bar'])
        self.assertEqual({'speed': '88mph',
                          'stores': 'bar,foo'},
                         self.image.extra_properties)

        # Addition adds to the list
        self.actions.merge_store_list('stores', ['baz'])
        self.assertEqual('bar,baz,foo', self.image.extra_properties['stores'])

        # Removal preserves the rest
        self.actions.merge_store_list('stores', ['foo'], subtract=True)
        self.assertEqual('bar,baz', self.image.extra_properties['stores'])

        # Duplicates aren't duplicated
        self.actions.merge_store_list('stores', ['bar'])
        self.assertEqual('bar,baz', self.image.extra_properties['stores'])

        # Removing the last store leaves the key empty but present
        self.actions.merge_store_list('stores', ['baz', 'bar'], subtract=True)
        self.assertEqual('', self.image.extra_properties['stores'])

        # Make sure we ignore falsey stores
        self.actions.merge_store_list('stores', ['', None])
        self.assertEqual('', self.image.extra_properties['stores'])

    @mock.patch.object(import_flow, 'LOG')
    def test_merge_store_logs_info(self, mock_log):
        # Removal from non-present key logs debug, but does not fail
        self.actions.merge_store_list('stores', ['foo,bar'], subtract=True)
        mock_log.debug.assert_has_calls([
            mock.call(
                'Stores %(stores)s not in %(key)s for image %(image_id)s',
                {'image_id': IMAGE_ID1,
                 'key': 'stores',
                 'stores': 'foo,bar'}),
            mock.call(
                'Image %(image_id)s %(key)s=%(stores)s',
                {'image_id': IMAGE_ID1,
                 'key': 'stores',
                 'stores': ''}),
        ])

        mock_log.debug.reset_mock()

        self.actions.merge_store_list('stores', ['foo'])
        self.assertEqual('foo', self.image.extra_properties['stores'])

        mock_log.debug.reset_mock()

        # Removal from a list where store is not present logs debug,
        # but does not fail
        self.actions.merge_store_list('stores', ['bar'], subtract=True)
        self.assertEqual('foo', self.image.extra_properties['stores'])
        mock_log.debug.assert_has_calls([
            mock.call(
                'Stores %(stores)s not in %(key)s for image %(image_id)s',
                {'image_id': IMAGE_ID1,
                 'key': 'stores',
                 'stores': 'bar'}),
            mock.call(
                'Image %(image_id)s %(key)s=%(stores)s',
                {'image_id': IMAGE_ID1,
                 'key': 'stores',
                 'stores': 'foo'}),
        ])

    def test_store_list_helpers(self):
        self.actions.add_importing_stores(['foo', 'bar', 'baz'])
        self.actions.remove_importing_stores(['bar'])
        self.actions.add_failed_stores(['foo', 'bar'])
        self.actions.remove_failed_stores(['foo'])
        self.assertEqual({'speed': '88mph',
                          'os_glance_importing_to_stores': 'baz,foo',
                          'os_glance_failed_import': 'bar'},
                         self.image.extra_properties)

    @mock.patch.object(image_import, 'set_image_data')
    def test_set_image_data(self, mock_sid):
        self.assertEqual(mock_sid.return_value,
                         self.actions.set_image_data(
                             mock.sentinel.uri, mock.sentinel.task_id,
                             mock.sentinel.backend, mock.sentinel.set_active))
        mock_sid.assert_called_once_with(
            self.image, mock.sentinel.uri, mock.sentinel.task_id,
            backend=mock.sentinel.backend, set_active=mock.sentinel.set_active,
            callback=None)

    @mock.patch.object(image_import, 'set_image_data')
    def test_set_image_data_with_callback(self, mock_sid):
        def fake_set_image_data(image, uri, task_id, backend=None,
                                set_active=False,
                                callback=None):
            callback(mock.sentinel.chunk, mock.sentinel.total)

        mock_sid.side_effect = fake_set_image_data

        callback = mock.MagicMock()
        self.actions.set_image_data(mock.sentinel.uri, mock.sentinel.task_id,
                                    mock.sentinel.backend,
                                    mock.sentinel.set_active,
                                    callback=callback)

        # Make sure our callback was triggered through the functools.partial
        # to include the original params and the action wrapper
        callback.assert_called_once_with(self.actions,
                                         mock.sentinel.chunk,
                                         mock.sentinel.total)

    def test_remove_location_for_store(self):
        self.image.locations = [
            {},
            {'metadata': {}},
            {'metadata': {'store': 'foo'}},
            {'metadata': {'store': 'bar'}},
        ]

        self.actions.remove_location_for_store('foo')
        self.assertEqual([{}, {'metadata': {}},
                          {'metadata': {'store': 'bar'}}],
                         self.image.locations)

        # Add a second definition for bar and make sure only one is removed
        self.image.locations.append({'metadata': {'store': 'bar'}})
        self.actions.remove_location_for_store('bar')
        self.assertEqual([{}, {'metadata': {}},
                          {'metadata': {'store': 'bar'}}],
                         self.image.locations)

    def test_remove_location_for_store_last_location(self):
        self.image.locations = [{'metadata': {'store': 'foo'}}]
        self.actions.remove_location_for_store('foo')
        self.assertEqual([], self.image.locations)
        self.assertIsNone(self.image.checksum)
        self.assertIsNone(self.image.os_hash_algo)
        self.assertIsNone(self.image.os_hash_value)
        self.assertIsNone(self.image.size)

    @mock.patch.object(import_flow, 'LOG')
    def test_remove_location_for_store_pop_failures(self, mock_log):
        class TestList(list):
            def pop(self):
                pass

        self.image.locations = TestList([{'metadata': {'store': 'foo'}}])
        with mock.patch.object(self.image.locations, 'pop',
                               new_callable=mock.PropertyMock) as mock_pop:

            mock_pop.side_effect = store_exceptions.NotFound(image='image')
            self.actions.remove_location_for_store('foo')
            mock_log.warning.assert_called_once_with(
                _('Error deleting from store foo when reverting.'))
            mock_log.warning.reset_mock()

            mock_pop.side_effect = store_exceptions.Forbidden()
            self.actions.remove_location_for_store('foo')
            mock_log.warning.assert_called_once_with(
                _('Error deleting from store foo when reverting.'))
            mock_log.warning.reset_mock()

            mock_pop.side_effect = Exception
            self.actions.remove_location_for_store('foo')
            mock_log.warning.assert_called_once_with(
                _('Unexpected exception when deleting from store foo.'))
            mock_log.warning.reset_mock()

    def test_pop_extra_property(self):
        self.image.extra_properties = {'foo': '1', 'bar': 2}

        # Should remove, if present
        self.actions.pop_extra_property('foo')
        self.assertEqual({'bar': 2}, self.image.extra_properties)

        # Should not raise if missing
        self.actions.pop_extra_property('baz')
        self.assertEqual({'bar': 2}, self.image.extra_properties)


@mock.patch('glance.common.scripts.utils.get_task')
class TestCompleteTask(test_utils.BaseTestCase):
    def setUp(self):
        super(TestCompleteTask, self).setUp()
        self.task_repo = mock.MagicMock()
        self.task = mock.MagicMock()
        self.wrapper = mock.MagicMock(image_id=IMAGE_ID1)

    def test_execute(self, mock_get_task):
        complete = import_flow._CompleteTask(TASK_ID1, TASK_TYPE,
                                             self.task_repo, self.wrapper)
        mock_get_task.return_value = self.task
        complete.execute()
        mock_get_task.assert_called_once_with(self.task_repo,
                                              TASK_ID1)
        self.task.succeed.assert_called_once_with({'image_id': IMAGE_ID1})
        self.task_repo.save.assert_called_once_with(self.task)
        self.wrapper.drop_lock_for_task.assert_called_once_with()

    def test_execute_no_task(self, mock_get_task):
        mock_get_task.return_value = None
        complete = import_flow._CompleteTask(TASK_ID1, TASK_TYPE,
                                             self.task_repo, self.wrapper)
        complete.execute()
        self.task_repo.save.assert_not_called()
        self.wrapper.drop_lock_for_task.assert_called_once_with()

    def test_execute_succeed_fails(self, mock_get_task):
        mock_get_task.return_value = self.task
        self.task.succeed.side_effect = Exception('testing')
        complete = import_flow._CompleteTask(TASK_ID1, TASK_TYPE,
                                             self.task_repo, self.wrapper)
        complete.execute()
        self.task.fail.assert_called_once_with(
            _('Error: <class \'Exception\'>: testing'))
        self.task_repo.save.assert_called_once_with(self.task)
        self.wrapper.drop_lock_for_task.assert_called_once_with()

    def test_execute_drop_lock_fails(self, mock_get_task):
        mock_get_task.return_value = self.task
        self.wrapper.drop_lock_for_task.side_effect = exception.NotFound()
        complete = import_flow._CompleteTask(TASK_ID1, TASK_TYPE,
                                             self.task_repo, self.wrapper)
        with mock.patch('glance.async_.flows.api_image_import.LOG') as m_log:
            complete.execute()
            m_log.error.assert_called_once_with('Image %(image)s import task '
                                                '%(task)s did not hold the '
                                                'lock upon completion!',
                                                {'image': IMAGE_ID1,
                                                 'task': TASK_ID1})
        self.task.succeed.assert_called_once_with({'image_id': IMAGE_ID1})


class TestImportMetadata(test_utils.BaseTestCase):
    def setUp(self):
        super(TestImportMetadata, self).setUp()
        self.config(extra_properties=[],
                    group="glance_download_properties")
        self.wrapper = mock.MagicMock(image_id=IMAGE_ID1)
        self.context = context.RequestContext(user_id=TENANT1,
                                              project_id=TENANT1,
                                              overwrite=False)
        self.import_req = {
            'method': {
                'glance_region': 'RegionTwo',
                'glance_service_interface': 'public',
                'glance_image_id': IMAGE_ID1
            }
        }

    @mock.patch('urllib.request')
    @mock.patch('glance.async_.flows.api_image_import.json')
    @mock.patch('glance.async_.utils.get_glance_endpoint')
    def test_execute_return_image_size(self, mock_gge, mock_json,
                                       mock_request):
        self.config(extra_properties=['hw:numa_nodes', 'os_hash'],
                    group="glance_download_properties")
        mock_gge.return_value = 'https://other.cloud.foo/image'
        action = self.wrapper.__enter__.return_value
        mock_json.loads.return_value = {
            'status': 'active',
            'disk_format': 'qcow2',
            'container_format': 'bare',
            'hw:numa_nodes': '2',
            'os_hash': 'hash',
            'extra_metadata': 'hello',
            'size': '12345'
        }
        mock_opener = mock.MagicMock()
        mock_payload = mock.MagicMock()
        mock_payload.read.return_value = b'{"status": "active"}'
        mock_opener.open.return_value.__enter__.return_value = mock_payload
        mock_request.build_opener.return_value = mock_opener
        task = import_flow._ImportMetadata(TASK_ID1, TASK_TYPE,
                                           self.context, self.wrapper,
                                           self.import_req)
        self.assertEqual(12345, task.execute())
        mock_request.Request.assert_called_once_with(
            'https://other.cloud.foo/image/v2/images/%s' % (
                IMAGE_ID1),
            headers={'X-Auth-Token': self.context.auth_token})
        mock_request.build_opener.assert_called_once()
        mock_gge.assert_called_once_with(self.context, 'RegionTwo', 'public')
        action.set_image_attribute.assert_called_once_with(
            disk_format='qcow2',
            container_format='bare')
        action.set_image_extra_properties.assert_called_once_with({
            'hw:numa_nodes': '2',
            'os_hash': 'hash'
        })

    @mock.patch('urllib.request')
    @mock.patch('glance.async_.utils.get_glance_endpoint')
    def test_execute_fail_no_glance_endpoint(self, mock_gge, mock_request):
        action = self.wrapper.__enter__.return_value
        mock_gge.side_effect = exception.GlanceEndpointNotFound(
            region='RegionTwo',
            interface='public')
        task = import_flow._ImportMetadata(TASK_ID1, TASK_TYPE,
                                           self.context, self.wrapper,
                                           self.import_req)
        self.assertRaises(exception.GlanceEndpointNotFound,
                          task.execute)
        action.assert_not_called()
        mock_request.assert_not_called()

    @mock.patch('urllib.request')
    @mock.patch('glance.async_.utils.get_glance_endpoint')
    def test_execute_fail_remote_glance_unreachable(self, mock_gge, mock_r):
        action = self.wrapper.__enter__.return_value
        mock_gge.return_value = 'https://other.cloud.foo/image'
        mock_opener = mock.MagicMock()
        mock_opener.open.side_effect = urllib.error.HTTPError(
            '/file', 400, 'Test Fail', {}, None)
        mock_r.build_opener.return_value = mock_opener
        task = import_flow._ImportMetadata(TASK_ID1, TASK_TYPE,
                                           self.context, self.wrapper,
                                           self.import_req)
        self.assertRaises(urllib.error.HTTPError,
                          task.execute)
        action.assert_not_called()

    @mock.patch('urllib.request')
    @mock.patch('glance.async_.flows.api_image_import.json')
    @mock.patch('glance.async_.utils.get_glance_endpoint')
    def test_execute_invalid_remote_image_state(self, mock_gge, mock_json,
                                                mock_request):
        action = self.wrapper.__enter__.return_value
        mock_gge.return_value = 'https://other.cloud.foo/image'
        mock_json.loads.return_value = {
            'status': 'queued',
        }
        mock_opener = mock.MagicMock()
        mock_payload = mock.MagicMock()
        mock_payload.read.return_value = b'{"status": "queued"}'
        mock_opener.open.return_value.__enter__.return_value = mock_payload
        mock_request.build_opener.return_value = mock_opener
        task = import_flow._ImportMetadata(TASK_ID1, TASK_TYPE,
                                           self.context, self.wrapper,
                                           self.import_req)
        self.assertRaises(import_flow._InvalidGlanceDownloadImageStatus,
                          task.execute)
        action.assert_not_called()

    @mock.patch('urllib.request')
    @mock.patch('glance.async_.flows.api_image_import.json')
    @mock.patch('glance.async_.utils.get_glance_endpoint')
    def test_execute_raise_if_no_size(self, mock_gge, mock_json, mock_request):
        self.config(extra_properties=['hw:numa_nodes', 'os_hash'],
                    group="glance_download_properties")
        mock_gge.return_value = 'https://other.cloud.foo/image'
        action = self.wrapper.__enter__.return_value
        mock_json.loads.return_value = {
            'status': 'active',
            'disk_format': 'qcow2',
            'container_format': 'bare',
            'hw:numa_nodes': '2',
            'os_hash': 'hash',
            'extra_metadata': 'hello',
        }
        mock_opener = mock.MagicMock()
        mock_payload = mock.MagicMock()
        mock_payload.read.return_value = b'{"status": "active"}'
        mock_opener.open.return_value.__enter__.return_value = mock_payload
        mock_request.build_opener.return_value = mock_opener
        task = import_flow._ImportMetadata(TASK_ID1, TASK_TYPE,
                                           self.context, self.wrapper,
                                           self.import_req)
        self.assertRaises(exception.ImportTaskError, task.execute)
        mock_request.Request.assert_called_once_with(
            'https://other.cloud.foo/image/v2/images/%s' % (
                IMAGE_ID1),
            headers={'X-Auth-Token': self.context.auth_token})
        mock_request.build_opener.assert_called_once()
        mock_gge.assert_called_once_with(self.context, 'RegionTwo', 'public')
        action.set_image_attribute.assert_called_once_with(
            disk_format='qcow2',
            container_format='bare')
        action.set_image_extra_properties.assert_called_once_with({
            'hw:numa_nodes': '2',
            'os_hash': 'hash'
        })

    @mock.patch('urllib.request')
    @mock.patch('glance.async_.utils.get_glance_endpoint')
    def test_import_metadata_redirect_validation(self, mock_gge,
                                                 mock_request):
        """Test redirect destinations are validated during metadata fetch."""
        mock_gge.return_value = 'https://other.cloud.foo/image'
        task = import_flow._ImportMetadata(TASK_ID1, TASK_TYPE,
                                           self.context, self.wrapper,
                                           self.import_req)
        mock_opener = mock.MagicMock()
        # Simulate redirect to disallowed URL
        mock_opener.open.side_effect = exception.ImportTaskError(
            "Redirect to disallowed URL: http://127.0.0.1:5000/")
        mock_request.build_opener.return_value = mock_opener
        self.assertRaises(exception.ImportTaskError, task.execute)
        # Verify SafeRedirectHandler is used
        mock_request.build_opener.assert_called_once()
        # Verify the handler passed is SafeRedirectHandler
        call_args = mock_request.build_opener.call_args
        # Check if SafeRedirectHandler class or instance is in args
        found_handler = (
            any(isinstance(arg, script_utils.SafeRedirectHandler)
                for arg in call_args.args) or
            script_utils.SafeRedirectHandler in call_args.args)
        self.assertTrue(
            found_handler,
            "SafeRedirectHandler should be used for redirect validation")

    @mock.patch('urllib.request')
    @mock.patch('glance.async_.flows.api_image_import.json')
    @mock.patch('glance.async_.utils.get_glance_endpoint')
    def test_import_metadata_uses_safe_redirect_handler(self, mock_gge,
                                                        mock_json,
                                                        mock_request):
        """Test that SafeRedirectHandler is used and allows valid redirects."""
        mock_gge.return_value = 'https://other.cloud.foo/image'
        mock_json.loads.return_value = {
            'status': 'active',
            'disk_format': 'qcow2',
            'container_format': 'bare',
            'size': '12345'
        }
        mock_opener = mock.MagicMock()
        mock_payload = mock.MagicMock()
        mock_payload.read.return_value = b'{"status": "active"}'
        mock_opener.open.return_value.__enter__.return_value = mock_payload
        mock_request.build_opener.return_value = mock_opener
        task = import_flow._ImportMetadata(TASK_ID1, TASK_TYPE,
                                           self.context, self.wrapper,
                                           self.import_req)
        # Execute should succeed with valid redirect
        result = task.execute()
        # Verify build_opener was called with SafeRedirectHandler
        mock_request.build_opener.assert_called_once()
        call_args = mock_request.build_opener.call_args
        found_handler = (
            any(isinstance(arg, script_utils.SafeRedirectHandler)
                for arg in call_args.args) or
            script_utils.SafeRedirectHandler in call_args.args)
        self.assertTrue(
            found_handler,
            "SafeRedirectHandler should be passed to build_opener")
        # Verify execution succeeded (handler allows valid redirects)
        self.assertEqual(12345, result)

    def test_revert_rollback_metadata_value(self):
        action = self.wrapper.__enter__.return_value
        task = import_flow._ImportMetadata(TASK_ID1, TASK_TYPE,
                                           self.context, self.wrapper,
                                           self.import_req)
        task.properties = {'prop1': 'value1', 'prop2': 'value2'}
        task.old_properties = {'prop1': 'orig_val', 'old_prop': 'old_value'}
        task.old_attributes = {'container_format': 'bare',
                               'disk_format': 'qcow2'}
        task.revert(None)
        action.set_image_attribute.assert_called_once_with(
            status='queued',
            container_format='bare',
            disk_format='qcow2')
        action.pop_extra_property.assert_called_once_with('prop2')
        action.set_image_extra_properties.assert_called_once_with(
            task.old_properties)



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_db import exception as db_exc
from oslo_log import log as logging
import sqlalchemy.exc as sa_exc
import sqlalchemy.orm as sa_orm

from glance.common import exception as exc
import glance.db.sqlalchemy.metadef_api.utils as metadef_utils
from glance.db.sqlalchemy import models_metadef as models

LOG = logging.getLogger(__name__)


def get(context, session, name):
    """Get a resource type, raise if not found"""

    try:
        query = session.query(models.MetadefResourceType).filter_by(name=name)
        resource_type = query.one()
    except sa_orm.exc.NoResultFound:
        LOG.debug("No metadata definition resource-type found with name %s",
                  name)
        raise exc.MetadefResourceTypeNotFound(resource_type_name=name)

    return resource_type.to_dict()


def get_all(context, session):
    """Get a list of all resource types"""

    query = session.query(models.MetadefResourceType)
    resource_types = query.all()

    resource_types_list = []
    for rt in resource_types:
        resource_types_list.append(rt.to_dict())

    return resource_types_list


def create(context, session, values):
    """Create a resource_type, raise if it already exists."""

    resource_type = models.MetadefResourceType()
    metadef_utils.drop_protected_attrs(models.MetadefResourceType, values)
    resource_type.update(values.copy())
    try:
        resource_type.save(session=session)
    except db_exc.DBDuplicateEntry:
        LOG.debug("Can not create the metadata definition resource-type. "
                  "A resource-type with name=%s already exists.",
                  resource_type.name)
        raise exc.MetadefDuplicateResourceType(
            resource_type_name=resource_type.name)

    return resource_type.to_dict()


def update(context, session, values):
    """Update a resource type, raise if not found"""

    name = values['name']
    metadef_utils.drop_protected_attrs(models.MetadefResourceType, values)
    db_rec = get(context, session, name)
    db_rec.update(values.copy())
    db_rec.save(session=session)

    return db_rec.to_dict()


def delete(context, session, name):
    """Delete a resource type or raise if not found or is protected"""

    db_rec = get(context, session, name)
    if db_rec.protected is True:
        LOG.debug("Delete forbidden. Metadata definition resource-type %s is a"
                  " seeded-system type and can not be deleted.", name)
        raise exc.ProtectedMetadefResourceTypeSystemDelete(
            resource_type_name=name)

    try:
        session.delete(db_rec)
        session.flush()
    except db_exc.DBError as e:
        if isinstance(e.inner_exception, sa_exc.IntegrityError):
            LOG.debug("Could not delete Metadata definition resource-type %s"
                      ". It still has content", name)
            raise exc.MetadefIntegrityError(
                record_type='resource-type', record_name=name)
        else:
            raise

    return db_rec.to_dict()



# Copyright 2024 RedHat Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import hashlib
import io
from unittest import mock

import glance_store as store
from oslo_config import cfg
from oslo_utils import units

import glance.async_.flows.location_import as import_flow
from glance.common import exception
from glance import context
import glance.tests.unit.utils as unit_test_utils
import glance.tests.utils as test_utils


CONF = cfg.CONF

BASE_URI = unit_test_utils.BASE_URI

TASK_TYPE = 'location_import'
TASK_ID1 = 'dbbe7231-020f-4311-87e1-5aaa6da56c02'
IMAGE_ID1 = '41f5b3b0-f54c-4cef-bd45-ce3e376a142f'
UUID1 = 'c80a1a6c-bd1f-41c5-90ee-81afedb1d58d'
TENANT1 = '6838eb7b-6ded-434a-882c-b344c77fe8df'


class TestCalculateHashTask(test_utils.BaseTestCase):
    def setUp(self):
        super(TestCalculateHashTask, self).setUp()
        self.task_repo = mock.MagicMock()
        self.task = mock.MagicMock()
        self.hash_task_input = {
            'image_id': IMAGE_ID1,
        }
        self.image_repo = mock.MagicMock()
        self.image = self.image_repo.get.return_value
        self.image.image_id = IMAGE_ID1
        self.image.disk_format = 'raw'
        self.image.container_format = 'bare'
        self.config(do_secure_hash=True)
        self.config(http_retries='3')
        self.context = context.RequestContext(user_id=TENANT1,
                                              project_id=TENANT1,
                                              overwrite=False)

    def test_execute_calculate_hash(self):
        self.loc_url = '%s/fake_location_1' % (BASE_URI)
        self.image.status = 'queued'
        hashing_algo = CONF.hashing_algorithm

        location_update = import_flow._UpdateLocationTask(
            TASK_ID1, TASK_TYPE, self.image_repo, IMAGE_ID1, self.loc_url,
            self.context)

        location_update.execute()
        self.assertEqual(1, self.image.locations.append.call_count)

        set_image_active = import_flow._SetImageToActiveTask(
            TASK_ID1, TASK_TYPE, self.image_repo, IMAGE_ID1)
        set_image_active.execute()
        self.assertEqual('active', self.image.status)

        hash_calculation = import_flow._CalculateHash(TASK_ID1, TASK_TYPE,
                                                      self.image_repo,
                                                      IMAGE_ID1,
                                                      hashing_algo)
        hash_calculation.execute()
        self.assertIsNotNone(self.image.checksum)
        self.assertIsNotNone(self.image.os_hash_algo)
        self.assertIsNotNone(self.image.os_hash_value)
        self.assertEqual('active', self.image.status)

    def test_hash_calculation_retry_count(self):
        hashing_algo = CONF.hashing_algorithm
        self.image.checksum = None
        self.image.os_hash_value = None
        hash_calculation = import_flow._CalculateHash(TASK_ID1, TASK_TYPE,
                                                      self.image_repo,
                                                      IMAGE_ID1,
                                                      hashing_algo,
                                                      status='importing')

        self.image.get_data.side_effect = IOError
        self.config(http_retries='10')
        expected_msg = ("Hash calculation failed for image .* data")
        self.assertRaisesRegex(import_flow._HashCalculationFailed,
                               expected_msg,
                               hash_calculation.execute)
        self.assertEqual(CONF.http_retries, self.image.get_data.call_count)
        self.assertEqual(CONF.hashing_algorithm, self.image.os_hash_algo)
        self.assertIsNone(self.image.checksum)
        self.assertIsNone(self.image.os_hash_value)

        hash_calculation.revert(None)
        self.assertIsNone(self.image.os_hash_algo)

    def test_execute_hash_calculation_fails_without_validation_data(self):
        self.loc_url = '%s/fake_location_1' % (BASE_URI)
        self.image.status = 'queued'
        self.hash_task_input.update(loc_url=self.loc_url)
        self.image.checksum = None
        self.image.os_hash_value = None

        location_update = import_flow._UpdateLocationTask(
            TASK_ID1, TASK_TYPE, self.image_repo, IMAGE_ID1, self.loc_url,
            self.context)

        location_update.execute()
        self.assertEqual(1, self.image.locations.append.call_count)

        # Since Image is mocked here, self.image.locations will not be
        # set hence setting it here to check that it's not popped out
        # even after CalculateHash failure
        self.image.locations = ['%s/fake_location_1' % (BASE_URI)]
        set_image_active = import_flow._SetImageToActiveTask(
            TASK_ID1, TASK_TYPE, self.image_repo, IMAGE_ID1)
        set_image_active.execute()
        self.assertEqual('active', self.image.status)

        hashing_algo = CONF.hashing_algorithm
        hash_calculation = import_flow._CalculateHash(TASK_ID1, TASK_TYPE,
                                                      self.image_repo,
                                                      IMAGE_ID1,
                                                      hashing_algo)

        self.image.get_data.side_effect = IOError
        with mock.patch.object(import_flow.LOG, 'debug') as mock_debug:
            hash_calculation.execute()
            debug_logs = mock_debug.call_args_list
            self.assertIn(("[%i/%i] Hash calculation failed due to %s",
                           1, 3, ''), debug_logs[0])
            self.assertEqual(CONF.hashing_algorithm, self.image.os_hash_algo)
            self.assertIsNone(self.image.checksum)
            self.assertIsNone(self.image.os_hash_value)
            self.assertEqual('active', self.image.status)
            self.assertEqual(1, len(self.image.locations))

        hash_calculation.revert(None)
        self.assertIsNone(self.image.os_hash_algo)
        self.assertEqual('active', self.image.status)
        self.assertEqual(1, len(self.image.locations))

        # Hash Calculation failed when image is 'active'.
        # exception will not be raised instead there will be warning log
        self.image.get_data.side_effect = IOError
        with mock.patch.object(import_flow.LOG, 'warning') as mock_warn:
            hash_calculation.execute()
            msg = ("Hash calculation failed for image %s data" % IMAGE_ID1)
            mock_warn.assert_called_once_with(msg)
            self.assertEqual(CONF.hashing_algorithm, self.image.os_hash_algo)
            self.assertIsNone(self.image.checksum)
            self.assertIsNone(self.image.os_hash_value)
            self.assertEqual('active', self.image.status)
            self.assertEqual(1, len(self.image.locations))

        hash_calculation.revert(None)
        self.assertIsNone(self.image.os_hash_algo)
        self.assertEqual('active', self.image.status)
        self.assertEqual(1, len(self.image.locations))

    def test_execute_hash_calculation_fails_for_store_other_that_http(self):
        self.loc_url = "cinder://image/fake_location"
        self.hash_task_input.update(loc_url=self.loc_url)
        self.image.status = 'queued'
        self.image.checksum = None
        self.image.os_hash_value = None

        location_update = import_flow._UpdateLocationTask(
            TASK_ID1, TASK_TYPE, self.image_repo, IMAGE_ID1, self.loc_url,
            self.context)

        location_update.execute()
        self.assertEqual(1, self.image.locations.append.call_count)

        # Since Image is mocked here, self.image.locations will not be
        # set hence setting it here to check that it's not popped out
        # even after CalculateHash failure
        self.image.locations = [{'url': 'cinder://image/fake_location'}]

        hashing_algo = CONF.hashing_algorithm
        hash_calculation = import_flow._CalculateHash(TASK_ID1, TASK_TYPE,
                                                      self.image_repo,
                                                      IMAGE_ID1,
                                                      hashing_algo,
                                                      status='importing')

        self.image.get_data.side_effect = IOError
        expected_msg = ("Hash calculation failed for image .* data")
        self.assertRaisesRegex(import_flow._HashCalculationFailed,
                               expected_msg,
                               hash_calculation.execute)
        self.assertEqual(CONF.hashing_algorithm, self.image.os_hash_algo)
        self.assertIsNone(self.image.checksum)
        self.assertIsNone(self.image.os_hash_value)
        self.assertEqual('importing', self.image.status)
        self.assertEqual(1, len(self.image.locations))

        hash_calculation.revert(None)
        self.assertIsNone(self.image.os_hash_algo)
        self.assertEqual('queued', self.image.status)
        self.assertEqual(0, len(self.image.locations))

    def test_execute_hash_calculation_fails_if_image_data_deleted(self):
        self.loc_url = '%s/fake_location_1' % (BASE_URI)
        self.image.status = 'queued'
        self.hash_task_input.update(loc_url=self.loc_url)
        self.image.checksum = None
        self.image.os_hash_value = None

        location_update = import_flow._UpdateLocationTask(
            TASK_ID1, TASK_TYPE, self.image_repo, IMAGE_ID1, self.loc_url,
            self.context)

        location_update.execute()
        self.assertEqual(1, self.image.locations.append.call_count)

        set_image_active = import_flow._SetImageToActiveTask(
            TASK_ID1, TASK_TYPE, self.image_repo, IMAGE_ID1)
        set_image_active.execute()
        self.assertEqual('active', self.image.status)

        hashing_algo = CONF.hashing_algorithm
        hash_calculation = import_flow._CalculateHash(TASK_ID1, TASK_TYPE,
                                                      self.image_repo,
                                                      IMAGE_ID1,
                                                      hashing_algo)
        self.image.get_data.side_effect = store.exceptions.NotFound
        hash_calculation.execute()
        # Check if Image delete and image_repo.delete has been called
        # if exception raised
        self.image.delete.assert_called_once()
        self.image_repo.remove.assert_called_once_with(self.image)


class TestVerifyValidationDataTask(test_utils.BaseTestCase):
    def setUp(self):
        super(TestVerifyValidationDataTask, self).setUp()
        self.task_repo = mock.MagicMock()
        self.task = mock.MagicMock()
        self.val_data_task_input = {
            'image_id': IMAGE_ID1,
        }
        self.image_repo = mock.MagicMock()
        self.image = self.image_repo.get.return_value
        self.image.image_id = IMAGE_ID1
        self.image.disk_format = 'raw'
        self.image.container_format = 'bare'
        self.config(do_secure_hash=True)

    def test_execute_with_valid_validation_data(self):
        url = '%s/fake_location_1' % BASE_URI
        self.image.status = 'queued'
        self.image.locations = {"url": url, "metadata": {"store": "foo"}}
        expected_size = 4 * units.Ki
        expected_data = b"*" * expected_size
        self.image.get_data.return_value = io.BytesIO(expected_data)
        hash_value = hashlib.sha512(expected_data).hexdigest()
        hashing_algo = CONF.hashing_algorithm
        self.image.checksum = None
        self.image.os_hash_value = None
        val_data = {
            'os_hash_algo': hashing_algo,
            'os_hash_value': hash_value
        }
        self.val_data_task_input.update(val_data=val_data)

        hash_calculation = import_flow._CalculateHash(TASK_ID1, TASK_TYPE,
                                                      self.image_repo,
                                                      IMAGE_ID1,
                                                      hashing_algo,
                                                      status='importing')
        hash_calculation.execute()

        self.image.os_hash_algo = val_data.get("os_hash_algo",
                                               hashing_algo)

        verify_validation_data = import_flow._VerifyValidationData(
            TASK_ID1, TASK_TYPE, self.image_repo, IMAGE_ID1, val_data)

        verify_validation_data.execute()
        self.assertEqual('sha512', self.image.os_hash_algo)
        self.assertEqual(hash_value, self.image.os_hash_value)
        self.assertEqual('importing', self.image.status)

        set_image_active = import_flow._SetImageToActiveTask(
            TASK_ID1, TASK_TYPE, self.image_repo, IMAGE_ID1)
        set_image_active.execute()
        self.assertEqual('active', self.image.status)

    def test_execute_with_os_hash_value_other_than_512(self):
        url = '%s/fake_location_1' % BASE_URI
        self.image.status = 'queued'
        self.image.locations = {"url": url, "metadata": {"store": "foo"}}
        expected_size = 4 * units.Ki
        expected_data = b"*" * expected_size
        self.image.get_data.return_value = io.BytesIO(expected_data)
        hash_value = hashlib.sha256(expected_data).hexdigest()
        hashing_algo = 'sha256'
        self.image.checksum = None
        self.image.os_hash_value = None
        val_data = {
            'os_hash_algo': 'sha256',
            'os_hash_value': hash_value
        }

        hash_calculation = import_flow._CalculateHash(TASK_ID1, TASK_TYPE,
                                                      self.image_repo,
                                                      IMAGE_ID1,
                                                      hashing_algo,
                                                      status='importing')
        hash_calculation.execute()

        self.val_data_task_input.update(val_data=val_data)

        verify_validation_data = import_flow._VerifyValidationData(
            TASK_ID1, TASK_TYPE, self.image_repo, IMAGE_ID1, val_data)

        verify_validation_data.execute()
        self.assertEqual('sha256', self.image.os_hash_algo)
        self.assertEqual(hash_value, self.image.os_hash_value)
        self.assertEqual('importing', self.image.status)

        set_image_active = import_flow._SetImageToActiveTask(
            TASK_ID1, TASK_TYPE, self.image_repo, IMAGE_ID1)
        set_image_active.execute()
        self.assertEqual('active', self.image.status)

    def test_execute_with_invalid_validation_data(self):
        url = '%s/fake_location_1' % BASE_URI
        self.image.status = 'queued'
        self.image.locations = [{"url": url, "metadata": {"store": "foo"}}]
        expected_size = 4 * units.Ki
        expected_data = b"*" * expected_size
        self.image.get_data.return_value = io.BytesIO(expected_data)
        hashing_algo = CONF.hashing_algorithm
        val_data = {
            'os_hash_algo': hashing_algo,
            'os_hash_value': hashlib.sha512(b'image_service').hexdigest()
        }
        hash_calculation = import_flow._CalculateHash(TASK_ID1, TASK_TYPE,
                                                      self.image_repo,
                                                      IMAGE_ID1,
                                                      hashing_algo,
                                                      status='importing')
        hash_calculation.execute()

        self.assertEqual('importing', self.image.status)
        self.assertEqual(1, len(self.image.locations))
        verify_validation_data = import_flow._VerifyValidationData(
            TASK_ID1, TASK_TYPE, self.image_repo, IMAGE_ID1,
            val_data)
        expected_msg = ("os_hash_value: .* not matched with actual "
                        "os_hash_value: .*")
        self.assertRaisesRegex(exception.InvalidParameterValue,
                               expected_msg,
                               verify_validation_data.execute)
        verify_validation_data.revert(None)
        self.assertIsNone(self.image.os_hash_algo)
        self.assertIsNone(self.image.os_hash_value)
        self.assertIsNone(self.image.checksum)
        self.assertEqual('queued', self.image.status)


class TestSetHashValuesTask(test_utils.BaseTestCase):
    def setUp(self):
        super(TestSetHashValuesTask, self).setUp()
        self.task_repo = mock.MagicMock()
        self.task = mock.MagicMock()
        self.hash_task_input = {
            'image_id': IMAGE_ID1,
        }
        self.image_repo = mock.MagicMock()
        self.image = self.image_repo.get.return_value
        self.image.image_id = IMAGE_ID1
        self.image.disk_format = 'raw'
        self.image.container_format = 'bare'

    def test_execute_with_valid_validation_data(self):
        url = '%s/fake_location_1' % BASE_URI
        self.image.status = 'queued'
        self.image.locations = {"url": url, "metadata": {"store": "foo"}}
        expected_size = 4 * units.Ki
        expected_data = b"*" * expected_size
        self.image.get_data.return_value = io.BytesIO(expected_data)
        hash_value = hashlib.sha512(expected_data).hexdigest()
        val_data = {
            'os_hash_algo': 'sha512',
            'os_hash_value': hash_value
        }
        self.hash_task_input.update(val_data=val_data)

        set_hash_data = import_flow._SetHashValues(
            TASK_ID1, TASK_TYPE, self.image_repo, IMAGE_ID1, val_data)

        set_hash_data.execute()
        self.assertEqual('sha512', self.image.os_hash_algo)
        self.assertEqual(hash_value, self.image.os_hash_value)
        self.assertEqual('queued', self.image.status)

        set_image_active = import_flow._SetImageToActiveTask(
            TASK_ID1, TASK_TYPE, self.image_repo, IMAGE_ID1)
        set_image_active.execute()
        self.assertEqual('active', self.image.status)


class TestUpdateLocationTask(test_utils.BaseTestCase):
    def setUp(self):
        super(TestUpdateLocationTask, self).setUp()
        self.task_repo = mock.MagicMock()
        self.task = mock.MagicMock()
        self.location_task_input = {
            'image_id': IMAGE_ID1,
        }
        self.image_repo = mock.MagicMock()
        self.image = self.image_repo.get.return_value
        self.image.image_id = IMAGE_ID1
        self.image.disk_format = 'raw'
        self.image.container_format = 'bare'
        self.context = context.RequestContext(user_id=TENANT1,
                                              project_id=TENANT1,
                                              overwrite=False)

    def test_execute_with_valid_location(self):
        self.loc_url = '%s/fake_location_1' % (BASE_URI)
        self.image.status = 'queued'
        self.location_task_input.update(loc_url=self.loc_url)

        location_update = import_flow._UpdateLocationTask(
            TASK_ID1, TASK_TYPE, self.image_repo, IMAGE_ID1, self.loc_url,
            self.context)

        location_update.execute()
        self.assertEqual(1, self.image.locations.append.call_count)

        set_image_active = import_flow._SetImageToActiveTask(
            TASK_ID1, TASK_TYPE, self.image_repo, IMAGE_ID1)
        set_image_active.execute()
        self.assertEqual('active', self.image.status)

    def test_execute_with_invalid_location(self):
        self.image.locations.append.side_effect = exception.BadStoreUri
        loc_url = 'bogus_url'
        self.image.status = 'queued'
        self.location_task_input.update(loc_url=loc_url)

        location_update = import_flow._UpdateLocationTask(
            TASK_ID1, TASK_TYPE, self.image_repo, IMAGE_ID1, loc_url,
            self.context)
        self.assertRaises(import_flow._InvalidLocation,
                          location_update.execute)
        self.assertEqual('queued', self.image.status)


class TestSetImageToActiveTask(test_utils.BaseTestCase):
    def setUp(self):
        super(TestSetImageToActiveTask, self).setUp()
        self.task_repo = mock.MagicMock()
        self.task = mock.MagicMock()
        self.set_status_task_input = {
            'image_id': IMAGE_ID1,
        }
        self.image_repo = mock.MagicMock()
        self.image = self.image_repo.get.return_value
        self.image.image_id = IMAGE_ID1
        self.image.disk_format = 'raw'
        self.image.container_format = 'bare'
        self.context = context.RequestContext(user_id=TENANT1,
                                              project_id=TENANT1,
                                              overwrite=False)

    def test_execute_set_image_to_active_state(self):
        self.loc_url = '%s/fake_location_1' % (BASE_URI)
        self.image.status = 'queued'
        self.set_status_task_input.update(loc_url=self.loc_url)

        location_update = import_flow._UpdateLocationTask(
            TASK_ID1, TASK_TYPE, self.image_repo, IMAGE_ID1, self.loc_url,
            self.context)

        location_update.execute()
        self.assertEqual(1, self.image.locations.append.call_count)
        self.assertEqual('queued', self.image.status)

        set_image_active = import_flow._SetImageToActiveTask(
            TASK_ID1, TASK_TYPE, self.image_repo, IMAGE_ID1)
        set_image_active.execute()
        self.assertEqual('active', self.image.status)

    def test_execute_set_image_to_active_state_failure(self):
        self.loc_url = '%s/fake_location_1' % (BASE_URI)
        self.image.status = 'queued'
        self.set_status_task_input.update(loc_url=self.loc_url)

        location_update = import_flow._UpdateLocationTask(
            TASK_ID1, TASK_TYPE, self.image_repo, IMAGE_ID1, self.loc_url,
            self.context)

        location_update.execute()
        self.assertEqual(1, self.image.locations.append.call_count)
        self.assertEqual('queued', self.image.status)

        # Test if image failed while saving to active state
        self.image_repo.save.side_effect = ValueError
        set_image_active = import_flow._SetImageToActiveTask(
            TASK_ID1, TASK_TYPE, self.image_repo, IMAGE_ID1)
        self.assertRaises(ValueError, set_image_active.execute)

        # Test revert where location added in previous task is popped
        # out incase of this task failure which didn't set image status
        # 'active'.
        self.image_repo.save.side_effect = None
        self.image.status = 'queued'
        set_image_active.revert(None)
        self.assertEqual(0, self.image.locations.pop.call_count)
        self.assertEqual('queued', self.image.status)



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Cache driver that uses xattr file tags and requires a filesystem
that has atimes set.

Assumptions
===========

1. Cache data directory exists on a filesystem that updates atime on
   reads ('noatime' should NOT be set)

2. Cache data directory exists on a filesystem that supports xattrs.
   This is optional, but highly recommended since it allows us to
   present ops with useful information pertaining to the cache, like
   human readable filenames and statistics.

3. `glance-prune` is scheduled to run as a periodic job via cron. This
    is needed to run the LRU prune strategy to keep the cache size
    within the limits set by the config file.


Cache Directory Notes
=====================

The image cache data directory contains the main cache path, where the
active cache entries and subdirectories for handling partial downloads
and errored-out cache images.

The layout looks like:

$image_cache_dir/
  entry1
  entry2
  ...
  incomplete/
  invalid/
  queue/
"""
from contextlib import contextmanager
import errno
import os
import stat
import time

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import encodeutils
from oslo_utils import excutils
from oslo_utils import fileutils
import xattr

from glance.common import exception
from glance.i18n import _, _LI
from glance.image_cache.drivers import base

LOG = logging.getLogger(__name__)

CONF = cfg.CONF


class Driver(base.Driver):

    """
    Cache driver that uses xattr file tags and requires a filesystem
    that has atimes set.
    """

    def configure(self):
        """
        Configure the driver to use the stored configuration options
        Any store that needs special configuration should implement
        this method. If the store was not able to successfully configure
        itself, it should raise `exception.BadDriverConfiguration`
        """
        # Here we set up the various file-based image cache paths
        # that we need in order to find the files in different states
        # of cache management.
        self.set_paths()

        # We do a quick attempt to write a user xattr to a temporary file
        # to check that the filesystem is even enabled to support xattrs
        image_cache_dir = self.base_dir
        fake_image_filepath = os.path.join(image_cache_dir, 'checkme')
        with open(fake_image_filepath, 'wb') as fake_file:
            fake_file.write(b"XXX")
            fake_file.flush()
        try:
            set_xattr(fake_image_filepath, 'hits', '1')
        except IOError as e:
            if e.errno == errno.EOPNOTSUPP:
                msg = (_("The device housing the image cache directory "
                         "%(image_cache_dir)s does not support xattr. It is"
                         " likely you need to edit your fstab and add the "
                         "user_xattr option to the appropriate line for the"
                         " device housing the cache directory.") %
                       {'image_cache_dir': image_cache_dir})
                LOG.error(msg)
                raise exception.BadDriverConfiguration(driver_name="xattr",
                                                       reason=msg)
        else:
            # Cleanup after ourselves...
            fileutils.delete_if_exists(fake_image_filepath)

    def get_cache_size(self):
        """
        Returns the total size in bytes of the image cache.
        """
        sizes = []
        for path in get_all_regular_files(self.base_dir):
            file_info = os.stat(path)
            sizes.append(file_info[stat.ST_SIZE])
        return sum(sizes)

    def get_hit_count(self, image_id):
        """
        Return the number of hits that an image has.

        :param image_id: Opaque image identifier
        """
        if not self.is_cached(image_id):
            return 0

        path = self.get_image_filepath(image_id)
        return int(get_xattr(path, 'hits', default=0))

    def get_cached_images(self):
        """
        Returns a list of records about cached images.
        """
        LOG.debug("Gathering cached image entries.")
        entries = []
        for path in get_all_regular_files(self.base_dir):
            image_id = os.path.basename(path)

            entry = {'image_id': image_id}
            file_info = os.stat(path)
            entry['last_modified'] = file_info[stat.ST_MTIME]
            entry['last_accessed'] = file_info[stat.ST_ATIME]
            entry['size'] = file_info[stat.ST_SIZE]
            entry['hits'] = self.get_hit_count(image_id)

            entries.append(entry)
        return entries

    def is_cached(self, image_id):
        """
        Returns True if the image with the supplied ID has its image
        file cached.

        :param image_id: Image ID
        """
        return os.path.exists(self.get_image_filepath(image_id))

    def is_cacheable(self, image_id):
        """
        Returns True if the image with the supplied ID can have its
        image file cached, False otherwise.

        :param image_id: Image ID
        """
        # Make sure we're not already cached or caching the image
        return not (self.is_cached(image_id) or
                    self.is_being_cached(image_id))

    def is_being_cached(self, image_id):
        """
        Returns True if the image with supplied id is currently
        in the process of having its image file cached.

        :param image_id: Image ID
        """
        path = self.get_image_filepath(image_id, 'incomplete')
        return os.path.exists(path)

    def is_queued(self, image_id):
        """
        Returns True if the image identifier is in our cache queue.
        """
        path = self.get_image_filepath(image_id, 'queue')
        return os.path.exists(path)

    def delete_all_cached_images(self):
        """
        Removes all cached image files and any attributes about the images
        """
        deleted = 0
        for path in get_all_regular_files(self.base_dir):
            delete_cached_file(path)
            deleted += 1
        return deleted

    def delete_cached_image(self, image_id):
        """
        Removes a specific cached image file and any attributes about the image

        :param image_id: Image ID
        """
        path = self.get_image_filepath(image_id)
        delete_cached_file(path)

    def delete_all_queued_images(self):
        """
        Removes all queued image files and any attributes about the images
        """
        files = [f for f in get_all_regular_files(self.queue_dir)]
        for file in files:
            fileutils.delete_if_exists(file)
        return len(files)

    def delete_queued_image(self, image_id):
        """
        Removes a specific queued image file and any attributes about the image

        :param image_id: Image ID
        """
        path = self.get_image_filepath(image_id, 'queue')
        fileutils.delete_if_exists(path)

    def get_least_recently_accessed(self):
        """
        Return a tuple containing the image_id and size of the least recently
        accessed cached file, or None if no cached files.
        """
        stats = []
        for path in get_all_regular_files(self.base_dir):
            file_info = os.stat(path)
            stats.append((file_info[stat.ST_ATIME],  # access time
                          file_info[stat.ST_SIZE],   # size in bytes
                          path))                     # absolute path

        if not stats:
            return None

        stats.sort()
        return os.path.basename(stats[0][2]), stats[0][1]

    @contextmanager
    def open_for_write(self, image_id):
        """
        Open a file for writing the image file for an image
        with supplied identifier.

        :param image_id: Image ID
        """
        incomplete_path = self.get_image_filepath(image_id, 'incomplete')

        def set_attr(key, value):
            set_xattr(incomplete_path, key, value)

        def commit():
            set_attr('hits', 0)

            final_path = self.get_image_filepath(image_id)
            LOG.debug("Fetch finished, moving "
                      "'%(incomplete_path)s' to '%(final_path)s'",
                      dict(incomplete_path=incomplete_path,
                           final_path=final_path))
            os.rename(incomplete_path, final_path)

            # Make sure that we "pop" the image from the queue...
            if self.is_queued(image_id):
                LOG.debug("Removing image '%s' from queue after "
                          "caching it.", image_id)
                fileutils.delete_if_exists(
                    self.get_image_filepath(image_id, 'queue'))

        def rollback(e):
            set_attr('error', encodeutils.exception_to_unicode(e))

            invalid_path = self.get_image_filepath(image_id, 'invalid')
            LOG.debug("Fetch of cache file failed (%(e)s), rolling back by "
                      "moving '%(incomplete_path)s' to "
                      "'%(invalid_path)s'",
                      {'e': encodeutils.exception_to_unicode(e),
                       'incomplete_path': incomplete_path,
                       'invalid_path': invalid_path})
            os.rename(incomplete_path, invalid_path)

        try:
            with open(incomplete_path, 'wb') as cache_file:
                yield cache_file
        except Exception as e:
            with excutils.save_and_reraise_exception():
                rollback(e)
        else:
            commit()
        finally:
            # if the generator filling the cache file neither raises an
            # exception, nor completes fetching all data, neither rollback
            # nor commit will have been called, so the incomplete file
            # will persist - in that case remove it as it is unusable
            # example: ^c from client fetch
            if os.path.exists(incomplete_path):
                rollback('incomplete fetch')

    @contextmanager
    def open_for_read(self, image_id):
        """
        Open and yield file for reading the image file for an image
        with supplied identifier.

        :param image_id: Image ID
        """
        path = self.get_image_filepath(image_id)
        with open(path, 'rb') as cache_file:
            yield cache_file
        path = self.get_image_filepath(image_id)
        inc_xattr(path, 'hits', 1)

    def queue_image(self, image_id):
        """
        This adds a image to be cache to the queue.

        If the image already exists in the queue or has already been
        cached, we return False, True otherwise

        :param image_id: Image ID
        """
        if self.is_cached(image_id):
            LOG.info(_LI("Not queueing image '%s'. Already cached."), image_id)
            return False

        if self.is_being_cached(image_id):
            LOG.info(_LI("Not queueing image '%s'. Already being "
                         "written to cache"), image_id)
            return False

        if self.is_queued(image_id):
            LOG.info(_LI("Not queueing image '%s'. Already queued."), image_id)
            return False

        path = self.get_image_filepath(image_id, 'queue')
        LOG.debug("Queueing image '%s'.", image_id)

        # Touch the file to add it to the queue
        with open(path, "w"):
            pass

        return True

    def get_queued_images(self):
        """
        Returns a list of image IDs that are in the queue. The
        list should be sorted by the time the image ID was inserted
        into the queue.
        """
        files = [f for f in get_all_regular_files(self.queue_dir)]
        items = []
        for path in files:
            mtime = os.path.getmtime(path)
            items.append((mtime, os.path.basename(path)))

        items.sort()
        return [image_id for (modtime, image_id) in items]

    def _reap_old_files(self, dirpath, entry_type, grace=None):
        now = time.time()
        reaped = 0
        for path in get_all_regular_files(dirpath):
            mtime = os.path.getmtime(path)
            age = now - mtime
            if not grace:
                LOG.debug("No grace period, reaping '%(path)s'"
                          " immediately", {'path': path})
                delete_cached_file(path)
                reaped += 1
            elif age > grace:
                LOG.debug("Cache entry '%(path)s' exceeds grace period, "
                          "(%(age)i s > %(grace)i s)",
                          {'path': path, 'age': age, 'grace': grace})
                delete_cached_file(path)
                reaped += 1

        LOG.info(_LI("Reaped %(reaped)s %(entry_type)s cache entries"),
                 {'reaped': reaped, 'entry_type': entry_type})
        return reaped

    def reap_invalid(self, grace=None):
        """Remove any invalid cache entries

        :param grace: Number of seconds to keep an invalid entry around for
                      debugging purposes. If None, then delete immediately.
        """
        return self._reap_old_files(self.invalid_dir, 'invalid', grace=grace)

    def reap_stalled(self, grace=None):
        """Remove any stalled cache entries

        :param grace: Number of seconds to keep an invalid entry around for
                      debugging purposes. If None, then delete immediately.
        """
        return self._reap_old_files(self.incomplete_dir, 'stalled',
                                    grace=grace)

    def clean(self, stall_time=None):
        """
        Delete any image files in the invalid directory and any
        files in the incomplete directory that are older than a
        configurable amount of time.
        """
        self.reap_invalid()

        if stall_time is None:
            stall_time = CONF.image_cache_stall_time

        self.reap_stalled(stall_time)


def get_all_regular_files(basepath):
    for fname in os.listdir(basepath):
        path = os.path.join(basepath, fname)
        if os.path.isfile(path):
            yield path


def delete_cached_file(path):
    LOG.debug("Deleting image cache file '%s'", path)
    fileutils.delete_if_exists(path)


def _make_namespaced_xattr_key(key, namespace='user'):
    """
    Create a fully-qualified xattr-key by including the intended namespace.

    Namespacing differs among OSes[1]:

        FreeBSD: user, system
        Linux: user, system, trusted, security
        MacOS X: not needed

    Mac OS X won't break if we include a namespace qualifier, so, for
    simplicity, we always include it.

    --
    [1] http://en.wikipedia.org/wiki/Extended_file_attributes
    """
    namespaced_key = ".".join([namespace, key])
    return namespaced_key


def get_xattr(path, key, **kwargs):
    """Return the value for a particular xattr

    If the key doesn't not exist, or xattrs aren't supported by the file
    system then a KeyError will be raised, that is, unless you specify a
    default using kwargs.
    """
    namespaced_key = _make_namespaced_xattr_key(key)
    try:
        return xattr.getxattr(path, namespaced_key)
    except IOError:
        if 'default' in kwargs:
            return kwargs['default']
        else:
            raise


def set_xattr(path, key, value):
    """Set the value of a specified xattr.

    If xattrs aren't supported by the file-system, we skip setting the value.
    """
    namespaced_key = _make_namespaced_xattr_key(key)
    if not isinstance(value, bytes):
        value = str(value).encode('utf-8')
    xattr.setxattr(path, namespaced_key, value)


def inc_xattr(path, key, n=1):
    """
    Increment the value of an xattr (assuming it is an integer).

    BEWARE, this code *does* have a RACE CONDITION, since the
    read/update/write sequence is not atomic.

    Since the use-case for this function is collecting stats--not critical--
    the benefits of simple, lock-free code out-weighs the possibility of an
    occasional hit not being counted.
    """
    count = int(get_xattr(path, key))
    count += n
    set_xattr(path, key, str(count))



# Copyright 2014 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from unittest import mock

import glance.common.scripts as scripts
from glance.common.scripts.image_import import main as image_import
import glance.tests.utils as test_utils


class TestScripts(test_utils.BaseTestCase):

    def setUp(self):
        super(TestScripts, self).setUp()

    def test_run_task(self):
        task_id = mock.ANY
        task_type = 'import'
        context = mock.ANY
        task_repo = mock.ANY
        image_repo = mock.ANY
        image_factory = mock.ANY

        with mock.patch.object(image_import, 'run') as mock_run:
            scripts.run_task(task_id, task_type, context, task_repo,
                             image_repo, image_factory)

        mock_run.assert_called_once_with(task_id, context, task_repo,
                                         image_repo, image_factory)



# Copyright 2012 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from cryptography import exceptions as crypto_exception
import glance_store as store
from unittest import mock
import urllib

from oslo_config import cfg
from oslo_policy import policy

from glance.async_.flows._internal_plugins import base_download
from glance.common import exception
from glance.common import store_utils
from glance.common import wsgi
import glance.context
import glance.db.simple.api as simple_db


CONF = cfg.CONF

UUID1 = 'c80a1a6c-bd1f-41c5-90ee-81afedb1d58d'
UUID2 = '971ec09a-8067-4bc8-a91f-ae3557f1c4c7'

TENANT1 = '6838eb7b-6ded-434a-882c-b344c77fe8df'
TENANT2 = '2c014f32-55eb-467d-8fcb-4bd706012f81'

USER1 = '54492ba0-f4df-4e4e-be62-27f4d76b29cf'
USER2 = '0b3b3006-cb76-4517-ae32-51397e22c754'
USER3 = '2hss8dkl-d8jh-88yd-uhs9-879sdjsd8skd'

BASE_URI = 'http://storeurl.com/container'


def sort_url_by_qs_keys(url):
    # NOTE(kragniz): this only sorts the keys of the query string of a url.
    # For example, an input of '/v2/tasks?sort_key=id&sort_dir=asc&limit=10'
    # returns '/v2/tasks?limit=10&sort_dir=asc&sort_key=id'. This is to prevent
    # non-deterministic ordering of the query string causing problems with unit
    # tests.

    parsed = urllib.parse.urlparse(url)
    queries = urllib.parse.parse_qsl(parsed.query, True)
    sorted_query = sorted(queries, key=lambda x: x[0])

    encoded_sorted_query = urllib.parse.urlencode(sorted_query, True)

    url_parts = (parsed.scheme, parsed.netloc, parsed.path,
                 parsed.params, encoded_sorted_query,
                 parsed.fragment)

    return urllib.parse.urlunparse(url_parts)


def get_fake_request(path='', method='POST', is_admin=False, user=USER1,
                     roles=None, headers=None, tenant=TENANT1):
    if roles is None:
        roles = ['member', 'reader']

    req = wsgi.Request.blank(path)
    req.method = method
    req.headers = {'x-openstack-request-id': 'my-req'}

    if headers is not None:
        req.headers.update(headers)

    kwargs = {
        'user': user,
        'tenant': tenant,
        'roles': roles,
        'is_admin': is_admin,
    }

    req.context = glance.context.RequestContext(**kwargs)
    return req


def enforcer_from_rules(unparsed_rules):
    rules = policy.Rules.from_dict(unparsed_rules)
    enforcer = glance.api.policy.Enforcer(
        suppress_deprecation_warnings=True)
    enforcer.set_rules(rules, overwrite=True)
    return enforcer


def fake_get_size_from_backend(uri, context=None):
    return 1


def fake_get_verifier(context, img_signature_certificate_uuid,
                      img_signature_hash_method, img_signature,
                      img_signature_key_type):
    verifier = mock.Mock()
    if (img_signature is not None and img_signature == 'VALID'):
        verifier.verify.return_value = None
    else:
        ex = crypto_exception.InvalidSignature()
        verifier.verify.side_effect = ex
    return verifier


def get_fake_context(user=USER1, tenant=TENANT1, roles=None, is_admin=False):
    if roles is None:
        roles = ['member']

    kwargs = {
        'user': user,
        'tenant': tenant,
        'roles': roles,
        'is_admin': is_admin,
    }

    context = glance.context.RequestContext(**kwargs)
    return context


class FakeDB(object):

    def __init__(self, initialize=True):
        self.reset()
        if initialize:
            self.init_db()

    @staticmethod
    def init_db():
        images = [
            {'id': UUID1, 'owner': TENANT1, 'status': 'queued',
             'locations': [{'url': '%s/%s' % (BASE_URI, UUID1),
                            'metadata': {}, 'status': 'queued'}],
             'disk_format': 'raw', 'container_format': 'bare'},
            {'id': UUID2, 'owner': TENANT1, 'status': 'queued',
             'disk_format': 'raw', 'container_format': 'bare'},
        ]
        [simple_db.image_create(None, image) for image in images]

        members = [
            {'image_id': UUID1, 'member': TENANT1, 'can_share': True},
            {'image_id': UUID1, 'member': TENANT2, 'can_share': False},
        ]
        [simple_db.image_member_create(None, member) for member in members]

        simple_db.image_tag_set_all(None, UUID1, ['ping', 'pong'])

    @staticmethod
    def reset():
        simple_db.reset()

    def __getattr__(self, key):
        return getattr(simple_db, key)


class FakeStoreUtils(object):
    def __init__(self, store_api):
        self.store_api = store_api

    def safe_delete_from_backend(self, context, id, location):
        try:
            del self.store_api.data[location['url']]
        except KeyError:
            pass

    def schedule_delayed_delete_from_backend(self, context, id, location):
        pass

    def delete_image_location_from_backend(self, context,
                                           image_id, location):
        if CONF.delayed_delete:
            self.schedule_delayed_delete_from_backend(context, image_id,
                                                      location)
        else:
            self.safe_delete_from_backend(context, image_id, location)

    def validate_external_location(self, uri):
        if uri and urllib.parse.urlparse(uri).scheme:
            return store_utils.validate_external_location(uri)
        else:
            return True


class FakeStoreAPI(object):
    def __init__(self, store_metadata=None):
        self.data = {
            '%s/%s' % (BASE_URI, UUID1): ('XXX', 3),
            '%s/fake_location' % (BASE_URI): ('YYY', 3)
        }
        self.acls = {}
        if store_metadata is None:
            self.store_metadata = {}
        else:
            self.store_metadata = store_metadata

    def create_stores(self):
        pass

    def set_acls(self, uri, public=False, read_tenants=None,
                 write_tenants=None, context=None):
        if read_tenants is None:
            read_tenants = []
        if write_tenants is None:
            write_tenants = []

        self.acls[uri] = {
            'public': public,
            'read': read_tenants,
            'write': write_tenants,
        }

    def get_from_backend(self, location, offset=0,
                         chunk_size=None, context=None):
        try:
            scheme = location[:location.find('/') - 1]
            if scheme == 'unknown':
                raise store.UnknownScheme(scheme=scheme)
            return self.data[location]
        except KeyError:
            raise store.NotFound(image=location)

    def get_size_from_backend(self, location, context=None):
        return self.get_from_backend(location, context=context)[1]

    def add_to_backend(self, conf, image_id, data, size,
                       scheme=None, context=None, verifier=None):
        store_max_size = 7
        current_store_size = 2
        for location in self.data.keys():
            if image_id in location:
                raise exception.Duplicate()
        if not size:
            # 'data' is a string wrapped in a LimitingReader|CooperativeReader
            # pipeline, so peek under the hood of those objects to get at the
            # string itself.
            size = len(data.data.fd)
        if (current_store_size + size) > store_max_size:
            raise exception.StorageFull()
        if context.user_id == USER2:
            raise exception.Forbidden()
        if context.user_id == USER3:
            raise exception.StorageWriteDenied()
        self.data[image_id] = (data, size)
        checksum = 'Z'
        return (image_id, size, checksum, self.store_metadata)

    def add_to_backend_with_multihash(
            self, conf, image_id, data, size, hashing_algo,
            scheme=None, context=None, verifier=None):
        store_max_size = 7
        current_store_size = 2
        for location in self.data.keys():
            if image_id in location:
                raise exception.Duplicate()
        if not size:
            # 'data' is a string wrapped in a LimitingReader|CooperativeReader
            # pipeline, so peek under the hood of those objects to get at the
            # string itself.
            size = len(data.data.fd._source)
        if (current_store_size + size) > store_max_size:
            raise exception.StorageFull()
        if context.user_id == USER2:
            raise exception.Forbidden()
        if context.user_id == USER3:
            raise exception.StorageWriteDenied()
        self.data[image_id] = (data, size)
        checksum = 'Z'
        multihash = 'ZZ'
        return (image_id, size, checksum, multihash, self.store_metadata)

    def check_location_metadata(self, val, key=''):
        store.check_location_metadata(val)

    def delete_from_backend(self, uri, context=None):
        pass


class FakeStoreAPIReader(FakeStoreAPI):
    """A store API that actually reads from the data pipe."""

    def add_to_backend_with_multihash(self, conf, image_id, data, size,
                                      hashing_algo, scheme=None, context=None,
                                      verifier=None):
        for chunk in data:
            pass

        return super(FakeStoreAPIReader, self).add_to_backend_with_multihash(
            conf, image_id, data, size, hashing_algo,
            scheme=scheme, context=context, verifier=verifier)


class FakePolicyEnforcer(object):
    def __init__(self, *_args, **kwargs):
        self.rules = {}

    def enforce(self, _ctxt, action, target=None, **kwargs):
        """Raise Forbidden if a rule for given action is set to false."""
        if self.rules.get(action) is False:
            raise exception.Forbidden()

    def set_rules(self, rules):
        self.rules = rules


class FakeNotifier(object):
    def __init__(self, *_args, **kwargs):
        self.log = []

    def _notify(self, event_type, payload, level):
        log = {
            'notification_type': level,
            'event_type': event_type,
            'payload': payload
        }
        self.log.append(log)

    def warn(self, event_type, payload):
        self._notify(event_type, payload, 'WARN')

    def info(self, event_type, payload):
        self._notify(event_type, payload, 'INFO')

    def error(self, event_type, payload):
        self._notify(event_type, payload, 'ERROR')

    def debug(self, event_type, payload):
        self._notify(event_type, payload, 'DEBUG')

    def critical(self, event_type, payload):
        self._notify(event_type, payload, 'CRITICAL')

    def get_logs(self):
        return self.log


class FakeGateway(object):
    def __init__(self, image_factory=None, image_member_factory=None,
                 image_repo=None, task_factory=None, task_repo=None):
        self.image_factory = image_factory
        self.image_member_factory = image_member_factory
        self.image_repo = image_repo
        self.task_factory = task_factory
        self.task_repo = task_repo

    def get_image_factory(self, context):
        return self.image_factory

    def get_image_member_factory(self, context):
        return self.image_member_factory

    def get_repo(self, context):
        return self.image_repo

    def get_task_factory(self, context):
        return self.task_factory

    def get_task_repo(self, context):
        return self.task_repo


class FakeTask(object):
    def __init__(self, task_id, type=None, status=None):
        self.task_id = task_id
        self.type = type
        self.message = None
        self.input = None
        self._status = status
        self._executor = None

    def success(self, result):
        self.result = result
        self._status = 'success'

    def fail(self, message):
        self.message = message
        self._status = 'failure'


class FakeBaseDownloadPlugin(base_download.BaseDownload):
    def execute(self):
        pass



#!/usr/bin/env python

# Copyright 2018 RedHat Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
A simple cache management utility for Glance.
"""

import argparse
import collections
import datetime
import functools
import os
import sys
import time
import uuid

from oslo_utils import encodeutils
import prettytable

# If ../glance/__init__.py exists, add ../ to Python search path, so that
# it will override what happens to be installed in /usr/(local/)lib/python...
possible_topdir = os.path.normpath(os.path.join(os.path.abspath(sys.argv[0]),
                                   os.pardir,
                                   os.pardir))
if os.path.exists(os.path.join(possible_topdir, 'glance', '__init__.py')):
    sys.path.insert(0, possible_topdir)

from glance.common import exception
import glance.image_cache.client
from glance.version import version_info as version


SUCCESS = 0
FAILURE = 1


def validate_input(func):
    """Decorator to enforce validation on input"""
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        if len(args[0].command) > 2:
            print("Please specify the ID of the image you wish for command "
                  "'%s' from the cache as the first and only "
                  "argument." % args[0].command[0])
            return FAILURE
        if len(args[0].command) == 2:
            image_id = args[0].command[1]
            try:
                image_id = uuid.UUID(image_id)
            except ValueError:
                print("Image ID '%s' is not a valid UUID." % image_id)
                return FAILURE

        return func(args[0], **kwargs)
    return wrapped


def catch_error(action):
    """Decorator to provide sensible default error handling for actions."""
    def wrap(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                ret = func(*args, **kwargs)
                return SUCCESS if ret is None else ret
            except exception.NotFound:
                options = args[0]
                print("Cache management middleware not enabled on host %s" %
                      options.host)
                return FAILURE
            except exception.Forbidden:
                print("Not authorized to make this request.")
                return FAILURE
            except Exception as e:
                options = args[0]
                if options.debug:
                    raise
                print("Failed to %s. Got error:" % action)
                pieces = encodeutils.exception_to_unicode(e).split('\n')
                for piece in pieces:
                    print(piece)
                return FAILURE

        return wrapper
    return wrap


@catch_error('show cached images')
def list_cached(args):
    """%(prog)s list-cached [options]

    List all images currently cached.
    """
    client = get_client(args)
    images = client.get_cached_images()
    if not images:
        print("No cached images.")
        return SUCCESS

    print("Found %d cached images..." % len(images))

    pretty_table = prettytable.PrettyTable(("ID",
                                            "Last Accessed (UTC)",
                                            "Last Modified (UTC)",
                                            "Size",
                                            "Hits"))
    pretty_table.align['Size'] = "r"
    pretty_table.align['Hits'] = "r"

    for image in images:
        last_accessed = image['last_accessed']
        if last_accessed == 0:
            last_accessed = "N/A"
        else:
            last_accessed = datetime.datetime.utcfromtimestamp(
                last_accessed).isoformat()

        pretty_table.add_row((
            image['image_id'],
            last_accessed,
            datetime.datetime.utcfromtimestamp(
                image['last_modified']).isoformat(),
            image['size'],
            image['hits']))

    print(pretty_table.get_string())
    return SUCCESS


@catch_error('show queued images')
def list_queued(args):
    """%(prog)s list-queued [options]

    List all images currently queued for caching.
    """
    client = get_client(args)
    images = client.get_queued_images()
    if not images:
        print("No queued images.")
        return SUCCESS

    print("Found %d queued images..." % len(images))

    pretty_table = prettytable.PrettyTable(("ID",))

    for image in images:
        pretty_table.add_row((image,))

    print(pretty_table.get_string())


@catch_error('queue the specified image for caching')
@validate_input
def queue_image(args):
    """%(prog)s queue-image <IMAGE_ID> [options]

    Queues an image for caching.
    """
    image_id = args.command[1]
    if (not args.force and
        not user_confirm("Queue image %(image_id)s for caching?" %
                         {'image_id': image_id}, default=False)):
        return SUCCESS

    client = get_client(args)
    client.queue_image_for_caching(image_id)

    if args.verbose:
        print("Queued image %(image_id)s for caching" %
              {'image_id': image_id})

    return SUCCESS


@catch_error('delete the specified cached image')
@validate_input
def delete_cached_image(args):
    """%(prog)s delete-cached-image <IMAGE_ID> [options]

    Deletes an image from the cache.
    """
    image_id = args.command[1]
    if (not args.force and
        not user_confirm("Delete cached image %(image_id)s?" %
                         {'image_id': image_id}, default=False)):
        return SUCCESS

    client = get_client(args)
    client.delete_cached_image(image_id)

    if args.verbose:
        print("Deleted cached image %(image_id)s" % {'image_id': image_id})

    return SUCCESS


@catch_error('Delete all cached images')
def delete_all_cached_images(args):
    """%(prog)s delete-all-cached-images [options]

    Remove all images from the cache.
    """
    if (not args.force and
            not user_confirm("Delete all cached images?", default=False)):
        return SUCCESS

    client = get_client(args)
    num_deleted = client.delete_all_cached_images()

    if args.verbose:
        print("Deleted %(num_deleted)s cached images" %
              {'num_deleted': num_deleted})

    return SUCCESS


@catch_error('delete the specified queued image')
@validate_input
def delete_queued_image(args):
    """%(prog)s delete-queued-image <IMAGE_ID> [options]

    Deletes an image from the cache.
    """
    image_id = args.command[1]
    if (not args.force and
        not user_confirm("Delete queued image %(image_id)s?" %
                         {'image_id': image_id}, default=False)):
        return SUCCESS

    client = get_client(args)
    client.delete_queued_image(image_id)

    if args.verbose:
        print("Deleted queued image %(image_id)s" % {'image_id': image_id})

    return SUCCESS


@catch_error('Delete all queued images')
def delete_all_queued_images(args):
    """%(prog)s delete-all-queued-images [options]

    Remove all images from the cache queue.
    """
    if (not args.force and
            not user_confirm("Delete all queued images?", default=False)):
        return SUCCESS

    client = get_client(args)
    num_deleted = client.delete_all_queued_images()

    if args.verbose:
        print("Deleted %(num_deleted)s queued images" %
              {'num_deleted': num_deleted})

    return SUCCESS


def get_client(options):
    """Return a new client object to a Glance server.

    specified by the --host and --port options
    supplied to the CLI
    """
    # Generate auth_url based on identity_api_version
    identity_version = env('OS_IDENTITY_API_VERSION', default='3')
    auth_url = options.os_auth_url
    if identity_version == '3' and "/v3" not in auth_url:
        auth_url = auth_url + "/v3"
    elif identity_version == '2' and "/v2" not in auth_url:
        auth_url = auth_url + "/v2.0"

    user_domain_id = options.os_user_domain_id
    if not user_domain_id:
        user_domain_id = options.os_domain_id
    project_domain_id = options.os_project_domain_id
    if not user_domain_id:
        project_domain_id = options.os_domain_id

    return glance.image_cache.client.get_client(
        host=options.host,
        port=options.port,
        username=options.os_username,
        password=options.os_password,
        project=options.os_project_name,
        user_domain_id=user_domain_id,
        project_domain_id=project_domain_id,
        auth_url=auth_url,
        auth_strategy=options.os_auth_strategy,
        auth_token=options.os_auth_token,
        region=options.os_region_name,
        insecure=options.insecure)


def env(*vars, **kwargs):
    """Search for the first defined of possibly many env vars.

    Returns the first environment variable defined in vars, or
    returns the default defined in kwargs.
    """
    for v in vars:
        value = os.environ.get(v)
        if value:
            return value
    return kwargs.get('default', '')


def print_help(args):
    """
    Print help specific to a command
    """
    command = lookup_command(args.command[1])
    print(command.__doc__ % {'prog': os.path.basename(sys.argv[0])})


def parse_args(parser):
    """Set up the CLI and config-file options that may be
    parsed and program commands.

    :param parser: The option parser
    """
    parser.add_argument('command', default='help', nargs='+',
                        help='The command to execute')
    parser.add_argument('-v', '--verbose', default=False, action="store_true",
                        help="Print more verbose output.")
    parser.add_argument('-d', '--debug', default=False, action="store_true",
                        help="Print debugging output.")
    parser.add_argument('-H', '--host', metavar="ADDRESS", default="0.0.0.0",
                        help="Address of Glance API host.")
    parser.add_argument('-p', '--port', dest="port", metavar="PORT",
                        type=int, default=9292,
                        help="Port the Glance API host listens on.")
    parser.add_argument('-k', '--insecure', dest="insecure",
                        default=False, action="store_true",
                        help='Explicitly allow glance to perform "insecure" '
                             "SSL (https) requests. The server's certificate "
                             "will not be verified against any certificate "
                             "authorities. This option should be used with "
                             "caution.")
    parser.add_argument('-f', '--force', dest="force",
                        default=False, action="store_true",
                        help="Prevent select actions from requesting "
                             "user confirmation.")

    parser.add_argument('--os-auth-token',
                        dest='os_auth_token',
                        default=env('OS_AUTH_TOKEN'),
                        help='Defaults to env[OS_AUTH_TOKEN].')
    parser.add_argument('-A', '--os_auth_token', '--auth_token',
                        dest='os_auth_token',
                        help=argparse.SUPPRESS)

    parser.add_argument('--os-username',
                        dest='os_username',
                        default=env('OS_USERNAME'),
                        help='Defaults to env[OS_USERNAME].')
    parser.add_argument('-I', '--os_username',
                        dest='os_username',
                        help=argparse.SUPPRESS)

    parser.add_argument('--os-password',
                        dest='os_password',
                        default=env('OS_PASSWORD'),
                        help='Defaults to env[OS_PASSWORD].')
    parser.add_argument('-K', '--os_password',
                        dest='os_password',
                        help=argparse.SUPPRESS)

    parser.add_argument('--os-region-name',
                        dest='os_region_name',
                        default=env('OS_REGION_NAME'),
                        help='Defaults to env[OS_REGION_NAME].')
    parser.add_argument('-R', '--os_region_name',
                        dest='os_region_name',
                        help=argparse.SUPPRESS)

    parser.add_argument('--os-project-id',
                        dest='os_project_id',
                        default=env('OS_PROJECT_ID'),
                        help='Defaults to env[OS_PROJECT_ID].')
    parser.add_argument('--os_project_id',
                        dest='os_project_id',
                        help=argparse.SUPPRESS)

    parser.add_argument('--os-project-name',
                        dest='os_project_name',
                        default=env('OS_PROJECT_NAME'),
                        help='Defaults to env[OS_PROJECT_NAME].')
    parser.add_argument('-T', '--os_project_name',
                        dest='os_project_name',
                        help=argparse.SUPPRESS)

    # arguments related user, project domain
    parser.add_argument('--os-user-domain-id',
                        dest='os_user_domain_id',
                        default=env('OS_USER_DOMAIN_ID'),
                        help='Defaults to env[OS_USER_DOMAIN_ID].')
    parser.add_argument('--os-project-domain-id',
                        dest='os_project_domain_id',
                        default=env('OS_PROJECT_DOMAIN_ID'),
                        help='Defaults to env[OS_PROJECT_DOMAIN_ID].')
    parser.add_argument('--os-domain-id',
                        dest='os_domain_id',
                        default=env('OS_DOMAIN_ID', default='default'),
                        help='Defaults to env[OS_DOMAIN_ID].')

    parser.add_argument('--os-auth-url',
                        default=env('OS_AUTH_URL'),
                        help='Defaults to env[OS_AUTH_URL].')
    parser.add_argument('-N', '--os_auth_url',
                        dest='os_auth_url',
                        help=argparse.SUPPRESS)

    parser.add_argument('-S', '--os_auth_strategy', dest="os_auth_strategy",
                        metavar="STRATEGY",
                        help="Authentication strategy (keystone or noauth).")

    version_string = version.cached_version_string()
    parser.add_argument('--version', action='version',
                        version=version_string)

    return parser.parse_args()


CACHE_COMMANDS = collections.OrderedDict()
CACHE_COMMANDS['help'] = (
    print_help, 'Output help for one of the commands below')
CACHE_COMMANDS['list-cached'] = (
    list_cached, 'List all images currently cached')
CACHE_COMMANDS['list-queued'] = (
    list_queued, 'List all images currently queued for caching')
CACHE_COMMANDS['queue-image'] = (
    queue_image, 'Queue an image for caching')
CACHE_COMMANDS['delete-cached-image'] = (
    delete_cached_image, 'Purges an image from the cache')
CACHE_COMMANDS['delete-all-cached-images'] = (
    delete_all_cached_images, 'Removes all images from the cache')
CACHE_COMMANDS['delete-queued-image'] = (
    delete_queued_image, 'Deletes an image from the cache queue')
CACHE_COMMANDS['delete-all-queued-images'] = (
    delete_all_queued_images, 'Deletes all images from the cache queue')


def _format_command_help():
    """Formats the help string for subcommands."""
    help_msg = "Commands:\n\n"

    for command, info in CACHE_COMMANDS.items():
        if command == 'help':
            command = 'help <command>'
        help_msg += "    %-28s%s\n\n" % (command, info[1])

    return help_msg


def lookup_command(command_name):
    try:
        command = CACHE_COMMANDS[command_name]
        return command[0]
    except KeyError:
        print('\nError: "%s" is not a valid command.\n' % command_name)
        print(_format_command_help())
        sys.exit("Unknown command: %(cmd_name)s" % {'cmd_name': command_name})


def user_confirm(prompt, default=False):
    """Yes/No question dialog with user.

    :param prompt: question/statement to present to user (string)
    :param default: boolean value to return if empty string
                    is received as response to prompt

    """
    if default:
        prompt_default = "[Y/n]"
    else:
        prompt_default = "[y/N]"

    answer = input("%s %s " % (prompt, prompt_default))

    if answer == "":
        return default
    else:
        return answer.lower() in ("yes", "y")


def main():
    print('In the Caracal development cycle, the glance-cache-manage command '
          'has been deprecated in favor of the new Cache API. It is scheduled '
          'to be removed in the Dalmatian development cycle.', file=sys.stderr)
    parser = argparse.ArgumentParser(
        description=_format_command_help(),
        formatter_class=argparse.RawDescriptionHelpFormatter)
    args = parse_args(parser)

    if args.command[0] == 'help' and len(args.command) == 1:
        parser.print_help()
        return

    # Look up the command to run
    command = lookup_command(args.command[0])

    try:
        start_time = time.time()
        result = command(args)
        end_time = time.time()
        if args.verbose:
            print("Completed in %-0.4f sec." % (end_time - start_time))
        sys.exit(result)
    except (RuntimeError, NotImplementedError) as e:
        sys.exit("ERROR: %s" % e)


if __name__ == '__main__':
    main()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_policy import policy

# Generic check string for checking if a user is authorized on a particular
# project, specifically with the member role.
PROJECT_MEMBER = 'role:member and project_id:%(project_id)s'
# Generic check string for checking if a user is authorized on a particular
# project but with read-only access. For example, this persona would be able to
# list private images owned by a project but cannot make any writeable changes
# to those images.
PROJECT_READER = 'role:reader and project_id:%(project_id)s'

# Make sure the member_id of the supplied target matches the project_id from
# the context object, which is derived from keystone tokens.
IMAGE_MEMBER_CHECK = 'project_id:%(member_id)s'
# Check if the visibility of the image supplied in the target matches
# "community"
COMMUNITY_VISIBILITY_CHECK = "'community':%(visibility)s"
# Check if the visibility of the resource supplied in the target matches
# "public"
PUBLIC_VISIBILITY_CHECK = "'public':%(visibility)s"
# Check if the visibility of the image supplied in the target matches "shared"
SHARED_VISIBILITY_CHECK = "'shared':%(visibility)s"

PROJECT_MEMBER_OR_IMAGE_MEMBER_OR_COMMUNITY_OR_PUBLIC_OR_SHARED = (
    f'role:member and (project_id:%(project_id)s or {IMAGE_MEMBER_CHECK} '
    f'or {COMMUNITY_VISIBILITY_CHECK} or {PUBLIC_VISIBILITY_CHECK} '
    f'or {SHARED_VISIBILITY_CHECK})'
)
PROJECT_READER_OR_IMAGE_MEMBER_OR_COMMUNITY_OR_PUBLIC_OR_SHARED = (
    f'role:reader and (project_id:%(project_id)s or {IMAGE_MEMBER_CHECK} '
    f'or {COMMUNITY_VISIBILITY_CHECK} or {PUBLIC_VISIBILITY_CHECK} '
    f'or {SHARED_VISIBILITY_CHECK})'
)
PROJECT_READER_OR_PUBLIC_NAMESPACE = (
    f'role:reader and (project_id:%(project_id)s or {PUBLIC_VISIBILITY_CHECK})'
)


# FIXME(lbragstad): These are composite check strings that represents glance's
# authorization code, some of which is implemented in the authorization wrapper
# and some is in the database driver.
#
# These check strings do not support tenancy with the `admin` role. This means
# anyone with the `admin` role on any project can execute a policy, which is
# typical in OpenStack services. But following check strings offer formal
# support for project membership and a read-only variant consistent with
# other OpenStack services.
ADMIN = 'rule:context_is_admin'
DEFAULT = 'rule:default'
ADMIN_OR_PROJECT_MEMBER = f'{ADMIN} or ({PROJECT_MEMBER})'
ADMIN_OR_PROJECT_READER = f'{ADMIN} or ({PROJECT_READER})'
ADMIN_OR_PROJECT_READER_GET_IMAGE = (
    f'{ADMIN} or '
    f'({PROJECT_READER_OR_IMAGE_MEMBER_OR_COMMUNITY_OR_PUBLIC_OR_SHARED})'
)
ADMIN_OR_PROJECT_MEMBER_DOWNLOAD_IMAGE = (
    f'{ADMIN} or '
    f'({PROJECT_MEMBER_OR_IMAGE_MEMBER_OR_COMMUNITY_OR_PUBLIC_OR_SHARED})'
)
ADMIN_OR_PROJECT_MEMBER_CREATE_IMAGE = (
    f'{ADMIN} or ({PROJECT_MEMBER} and project_id:%(owner)s)'
)
ADMIN_OR_PROJECT_READER_GET_NAMESPACE = (
    f'{ADMIN} or ({PROJECT_READER_OR_PUBLIC_NAMESPACE})'
)


ADMIN_OR_SHARED_MEMBER = (
    f'{ADMIN} or (role:member and {IMAGE_MEMBER_CHECK})'
)
ADMIN_OR_PROJECT_READER_OR_SHARED_MEMBER = (
    f'{ADMIN} or '
    f'role:reader and (project_id:%(project_id)s or {IMAGE_MEMBER_CHECK})'
)
SERVICE_OR_PROJECT_MEMBER = (
    f'rule:service_api or ({PROJECT_MEMBER} and project_id:%(owner)s)'
)

SERVICE = 'rule:service_api'

rules = [
    policy.RuleDefault(name='default', check_str='',
                       description='Defines the default rule used for '
                                   'policies that historically had an empty '
                                   'policy in the supplied policy.json file.',
                       deprecated_rule=policy.DeprecatedRule(
                           name='default',
                           check_str=ADMIN,
                           deprecated_reason='In order to allow operators to '
                           'accept the default policies from code by not '
                           'defining them in the policy file, while still '
                           'working with old policy files that rely on the '
                           '``default`` rule for policies that are '
                           'not specified in the policy file, the ``default`` '
                           'rule must now be explicitly set to '
                           '``"role:admin"`` when that is the desired default '
                           'for unspecified rules.',
                           deprecated_since='Ussuri')),
    policy.RuleDefault(name='context_is_admin', check_str='role:admin',
                       description='Defines the rule for the is_admin:True '
                                   'check.'),
    policy.RuleDefault(name='service_api', check_str='role:service',
                       description='Default rule for the service-to-service '
                       'API.'),
]


def list_rules():
    return rules



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
SQLAlchemy models for glance metadata schema
"""

from oslo_db.sqlalchemy import models
from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import ForeignKey
from sqlalchemy import Index
from sqlalchemy import Integer
from sqlalchemy.orm import declarative_base
from sqlalchemy.orm import relationship
from sqlalchemy import String
from sqlalchemy import Text
from sqlalchemy import UniqueConstraint

from glance.common import timeutils
from glance.db.sqlalchemy.models import JSONEncodedDict


class DictionaryBase(models.ModelBase):
    metadata = None

    def to_dict(self):
        d = {}
        for c in self.__table__.columns:
            d[c.name] = self[c.name]
        return d


BASE_DICT = declarative_base(cls=DictionaryBase)


class GlanceMetadefBase(models.TimestampMixin):
    """Base class for Glance Metadef Models."""

    __table_args__ = {'mysql_engine': 'InnoDB', 'mysql_charset': 'utf8'}
    __table_initialized__ = False
    __protected_attributes__ = set(["created_at", "updated_at"])

    created_at = Column(DateTime, default=lambda: timeutils.utcnow(),
                        nullable=False)
    # TODO(wko): Column `updated_at` have no default value in
    #            OpenStack common code. We should decide, is this value
    #            required and make changes in oslo (if required) or
    #            in glance (if not).
    updated_at = Column(DateTime, default=lambda: timeutils.utcnow(),
                        nullable=True, onupdate=lambda: timeutils.utcnow())


class MetadefNamespace(BASE_DICT, GlanceMetadefBase):
    """Represents a metadata-schema namespace in the datastore."""
    __tablename__ = 'metadef_namespaces'
    __table_args__ = (UniqueConstraint('namespace',
                                       name='uq_metadef_namespaces'
                                            '_namespace'),
                      Index('ix_metadef_namespaces_owner', 'owner')
                      )

    id = Column(Integer, primary_key=True, nullable=False)
    namespace = Column(String(80), nullable=False)
    display_name = Column(String(80))
    description = Column(Text())
    visibility = Column(String(32))
    protected = Column(Boolean)
    owner = Column(String(255), nullable=False)


class MetadefObject(BASE_DICT, GlanceMetadefBase):
    """Represents a metadata-schema object in the datastore."""
    __tablename__ = 'metadef_objects'
    __table_args__ = (UniqueConstraint('namespace_id', 'name',
                                       name='uq_metadef_objects_namespace_id'
                                            '_name'),
                      Index('ix_metadef_objects_name', 'name')
                      )

    id = Column(Integer, primary_key=True, nullable=False)
    namespace_id = Column(Integer(), ForeignKey('metadef_namespaces.id'),
                          nullable=False)
    name = Column(String(80), nullable=False)
    description = Column(Text())
    required = Column(Text())
    json_schema = Column(JSONEncodedDict(), default={}, nullable=False)


class MetadefProperty(BASE_DICT, GlanceMetadefBase):
    """Represents a metadata-schema namespace-property in the datastore."""
    __tablename__ = 'metadef_properties'
    __table_args__ = (UniqueConstraint('namespace_id', 'name',
                                       name='uq_metadef_properties_namespace'
                                            '_id_name'),
                      Index('ix_metadef_properties_name', 'name')
                      )

    id = Column(Integer, primary_key=True, nullable=False)
    namespace_id = Column(Integer(), ForeignKey('metadef_namespaces.id'),
                          nullable=False)
    name = Column(String(80), nullable=False)
    json_schema = Column(JSONEncodedDict(), default={}, nullable=False)


class MetadefNamespaceResourceType(BASE_DICT, GlanceMetadefBase):
    """Represents a metadata-schema namespace-property in the datastore."""
    __tablename__ = 'metadef_namespace_resource_types'
    __table_args__ = (Index('ix_metadef_ns_res_types_namespace_id',
                            'namespace_id'),
                      )

    resource_type_id = Column(Integer,
                              ForeignKey('metadef_resource_types.id'),
                              primary_key=True, nullable=False)
    namespace_id = Column(Integer, ForeignKey('metadef_namespaces.id'),
                          primary_key=True, nullable=False)
    properties_target = Column(String(80))
    prefix = Column(String(80))


class MetadefResourceType(BASE_DICT, GlanceMetadefBase):
    """Represents a metadata-schema resource type in the datastore."""
    __tablename__ = 'metadef_resource_types'
    __table_args__ = (UniqueConstraint('name',
                                       name='uq_metadef_resource_types_name'),
                      )

    id = Column(Integer, primary_key=True, nullable=False)
    name = Column(String(80), nullable=False)
    protected = Column(Boolean, nullable=False, default=False)

    associations = relationship(
        "MetadefNamespaceResourceType",
        primaryjoin=id == MetadefNamespaceResourceType.resource_type_id)


class MetadefTag(BASE_DICT, GlanceMetadefBase):
    """Represents a metadata-schema tag in the data store."""
    __tablename__ = 'metadef_tags'
    __table_args__ = (UniqueConstraint('namespace_id', 'name',
                                       name='uq_metadef_tags_namespace_id'
                                            '_name'),
                      Index('ix_metadef_tags_name', 'name')
                      )

    id = Column(Integer, primary_key=True, nullable=False)
    namespace_id = Column(Integer(), ForeignKey('metadef_namespaces.id'),
                          nullable=False)
    name = Column(String(80), nullable=False)


def register_models(engine):
    """Create database tables for all models with the given engine."""
    models = (MetadefNamespace, MetadefObject, MetadefProperty,
              MetadefTag,
              MetadefResourceType, MetadefNamespaceResourceType)
    for model in models:
        model.metadata.create_all(engine)


def unregister_models(engine):
    """Drop database tables for all models with the given engine."""
    models = (MetadefObject, MetadefProperty, MetadefNamespaceResourceType,
              MetadefTag,
              MetadefNamespace, MetadefResourceType)
    for model in models:
        model.metadata.drop_all(engine)



# Copyright 2021 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import oslo_policy.policy

from glance.api import policy
from glance.tests import functional


NAME_SPACE1 = {
    "namespace": "MyNamespace",
    "display_name": "My User Friendly Namespace",
    "description": "My description"
}

RESOURCETYPE_1 = {
    "name": "MyResourceType",
    "prefix": "prefix_",
    "properties_target": "temp"
}

RESOURCETYPE_2 = {
    "name": "MySecondResourceType",
    "prefix": "temp_prefix_",
    "properties_target": "temp_2"
}


class TestMetadefResourceTypesPolicy(functional.SynchronousAPIBase):
    def setUp(self):
        super(TestMetadefResourceTypesPolicy, self).setUp()
        self.policy = policy.Enforcer(suppress_deprecation_warnings=True)

    def load_data(self, create_resourcetypes=False):
        path = '/v2/metadefs/namespaces'
        md_resource = self._create_metadef_resource(path=path,
                                                    data=NAME_SPACE1)
        self.assertEqual('MyNamespace', md_resource['namespace'])
        if create_resourcetypes:
            namespace = md_resource['namespace']
            path = '/v2/metadefs/namespaces/%s/resource_types' % namespace
            for resource in [RESOURCETYPE_1, RESOURCETYPE_2]:
                md_resource = self._create_metadef_resource(path=path,
                                                            data=resource)
                self.assertEqual(resource['name'], md_resource['name'])

    def set_policy_rules(self, rules):
        self.policy.set_rules(
            oslo_policy.policy.Rules.from_dict(rules),
            overwrite=True)

    def start_server(self):
        with mock.patch.object(policy, 'Enforcer') as mock_enf:
            mock_enf.return_value = self.policy
            super(TestMetadefResourceTypesPolicy, self).start_server()

    def _verify_forbidden_converted_to_not_found(self, path, method,
                                                 json=None):
        # Note for other reviewers, these tests runs by default using
        # admin role, to test this scenario we need private namespace
        # of current project to be accessed by other projects non-admin
        # user.
        headers = self._headers({
            'X-Tenant-Id': 'fake-tenant-id',
            'X-Roles': 'member',
        })
        resp = self.api_request(method, path, headers=headers, json=json)
        self.assertEqual(404, resp.status_code)

    def test_namespace_resourcetypes_list_basic(self):
        self.start_server()
        # Create namespace and resourcetypes
        self.load_data(create_resourcetypes=True)
        # First make sure list resourcetypes works with default policy
        namespace = NAME_SPACE1['namespace']
        path = '/v2/metadefs/namespaces/%s/resource_types' % namespace
        resp = self.api_get(path)
        md_resource = resp.json
        self.assertEqual(2, len(md_resource['resource_type_associations']))

        # Now disable list_metadef_resource_types permissions and make
        # sure any other attempts fail
        self.set_policy_rules({
            'list_metadef_resource_types': '!',
            'get_metadef_namespace': '@',
        })
        resp = self.api_get(path)
        self.assertEqual(403, resp.status_code)

        # Now disable both permissions and make sure you will get
        # 404 Not Found
        self.set_policy_rules({
            'list_metadef_resource_types': '!',
            'get_metadef_namespace': '!',
        })
        resp = self.api_get(path)
        # Note for reviewers, this causes our "check get if list fails"
        # logic to return 404 as we expect, but not related to the latest
        # rev that checks the namespace get operation first.
        self.assertEqual(404, resp.status_code)

        # Now enable list_metadef_resource_types and disable
        # get_metadef_resource_type permission to make sure that you will get
        # empty list as a response
        self.set_policy_rules({
            'list_metadef_resource_types': '@',
            'get_metadef_resource_type': '!',
            'get_metadef_namespace': '@',
        })
        resp = self.api_get(path)
        md_resource = resp.json
        self.assertEqual(0, len(md_resource['resource_type_associations']))

        # Ensure accessing non visible namespace will catch 403 and
        # return 404 to user
        self.set_policy_rules({
            'list_metadef_resource_types': '@',
            'get_metadef_resource_type': '@',
            'get_metadef_namespace': '@',
        })
        self._verify_forbidden_converted_to_not_found(path, 'GET')

    def test_resourcetypes_list_basic(self):
        self.start_server()
        # Create namespace and resourcetypes
        self.load_data(create_resourcetypes=True)
        # First make sure list resourcetypes works with default policy
        path = '/v2/metadefs/resource_types'
        resp = self.api_get(path)
        md_resource = resp.json
        # NOTE(abhishekk): /v2/metadefs/resource_types returns list which
        # contains all resource_types in a dictionary, so the length will
        # always be 1 here.
        self.assertEqual(1, len(md_resource))

        # Now disable get_metadef_resource_type permissions and make
        # sure any other attempts fail
        self.set_policy_rules({
            'list_metadef_resource_types': '!',
            'get_metadef_namespace': '@',
        })
        resp = self.api_get(path)
        self.assertEqual(403, resp.status_code)

    def test_resourcetype_create_basic(self):
        self.start_server()
        # Create namespace
        self.load_data()
        # First make sure create resourcetype works with default policy
        namespace = NAME_SPACE1['namespace']
        path = '/v2/metadefs/namespaces/%s/resource_types' % namespace
        md_resource = self._create_metadef_resource(path=path,
                                                    data=RESOURCETYPE_1)
        self.assertEqual('MyResourceType', md_resource['name'])

        # Now disable add_metadef_resource_type_association permissions
        # and make sure any other attempts fail
        self.set_policy_rules({
            'add_metadef_resource_type_association': '!',
            'get_metadef_namespace': '@',
        })
        resp = self.api_post(path, json=RESOURCETYPE_2)
        self.assertEqual(403, resp.status_code)

        # Now disable both permissions and make sure you will get
        # 404 Not Found
        self.set_policy_rules({
            'add_metadef_resource_type_association': '!',
            'get_metadef_namespace': '!',
        })
        resp = self.api_post(path, json=RESOURCETYPE_2)
        # Note for reviewers, this causes our "check get if create fails"
        # logic to return 404 as we expect, but not related to the latest
        # rev that checks the namespace get operation first.
        self.assertEqual(404, resp.status_code)

        # Ensure accessing non visible namespace will catch 403 and
        # return 404 to user
        self.set_policy_rules({
            'add_metadef_resource_type_association': '@',
            'get_metadef_namespace': '@',
        })
        self._verify_forbidden_converted_to_not_found(path, 'POST',
                                                      json=RESOURCETYPE_2)

    def test_object_delete_basic(self):
        self.start_server()
        # Create namespace and objects
        self.load_data(create_resourcetypes=True)
        # Now ensure you are able to delete the resource_types
        path = '/v2/metadefs/namespaces/%s/resource_types/%s' % (
            NAME_SPACE1['namespace'], RESOURCETYPE_1['name'])
        resp = self.api_delete(path)
        self.assertEqual(204, resp.status_code)

        # Verify that resource_type is deleted
        namespace = NAME_SPACE1['namespace']
        path = '/v2/metadefs/namespaces/%s/resource_types' % namespace
        resp = self.api_get(path)
        md_resource = resp.json
        # assert namespace has only one resource type association
        self.assertEqual(1, len(md_resource['resource_type_associations']))
        # assert deleted association is not present in response
        for resource in md_resource['resource_type_associations']:
            self.assertNotEqual(RESOURCETYPE_1['name'], resource['name'])

        # Now disable remove_metadef_resource_type_association permissions
        # and make sure any other attempts fail
        path = '/v2/metadefs/namespaces/%s/resource_types/%s' % (
            NAME_SPACE1['namespace'], RESOURCETYPE_2['name'])
        self.set_policy_rules({
            'remove_metadef_resource_type_association': '!',
            'get_metadef_namespace': '@',
        })
        resp = self.api_delete(path)
        self.assertEqual(403, resp.status_code)

        # Now disable both permissions and make sure you will get
        # 404 Not Found
        self.set_policy_rules({
            'remove_metadef_resource_type_association': '!',
            'get_metadef_namespace': '!',
        })
        resp = self.api_delete(path)
        # Note for reviewers, this causes our "check get if delete fails"
        # logic to return 404 as we expect, but not related to the latest
        # rev that checks the namespace get operation first.
        self.assertEqual(404, resp.status_code)

        # Ensure accessing non visible namespace will catch 403 and
        # return 404 to user
        self.set_policy_rules({
            'remove_metadef_resource_type_association': '@',
            'get_metadef_namespace': '@',
        })
        self._verify_forbidden_converted_to_not_found(path, 'DELETE')
